from typing import Optional

import mlx.core as mx
import mlx.nn as nn

from ..base import (
    LanguageModelOutput,
    create_attention_mask,
    scaled_dot_product_attention,
)
from ..cache import KVCache
from .config import TextConfig


def _activation_from_config(config: TextConfig):
    act = getattr(config, "hidden_activation", None)
    if act is None:
        act = getattr(config, "hidden_act", "gelu_pytorch_tanh")

    if act == "gelu":
        return nn.GELU()
    if act in ("gelu_pytorch_tanh", "gelu_new", "gelu_fast", "tanh"):
        return nn.GELU(approx="precise")

    raise ValueError(f"Unsupported activation '{act}' for PaliGemma")


class RMSNorm(nn.Module):
    def __init__(self, dims: int, eps: float = 1e-6):
        super().__init__()
        self.weight = mx.ones((dims,))
        self.eps = eps

    def __call__(self, x):
        return mx.fast.rms_norm(x, 1.0 + self.weight, self.eps)


class Attention(nn.Module):
    def __init__(self, config: TextConfig):
        super().__init__()

        dim = config.hidden_size
        self.n_heads = n_heads = config.num_attention_heads
        self.n_kv_heads = n_kv_heads = config.num_key_value_heads
        self.model_type = config.model_type
        self.attn_logit_softcapping = config.attn_logit_softcapping
        self.repeats = n_heads // n_kv_heads
        self.head_dim = head_dim = (
            config.hidden_size // n_heads
            if self.model_type == "gemma"
            else config.head_dim
        )
        self.scale = (
            head_dim**-0.5
            if self.model_type == "gemma"
            else 1.0 / (config.query_pre_attn_scalar**0.5)
        )
        attention_bias = getattr(config, "attention_bias", False)

        self.q_proj = nn.Linear(dim, n_heads * head_dim, bias=attention_bias)
        self.k_proj = nn.Linear(dim, n_kv_heads * head_dim, bias=attention_bias)
        self.v_proj = nn.Linear(dim, n_kv_heads * head_dim, bias=attention_bias)
        self.o_proj = nn.Linear(n_heads * head_dim, dim, bias=attention_bias)

        self.rope = nn.RoPE(
            head_dim,
            traditional=config.rope_traditional,
            base=config.rope_theta,
        )

    def __call__(
        self,
        x: mx.array,
        mask: Optional[mx.array] = None,
        cache: Optional[KVCache] = None,
    ) -> mx.array:
        B, L, D = x.shape

        queries, keys, values = self.q_proj(x), self.k_proj(x), self.v_proj(x)

        # Prepare the queries, keys and values for the attention computation
        queries = queries.reshape(B, L, self.n_heads, -1).transpose(0, 2, 1, 3)
        keys = keys.reshape(B, L, self.n_kv_heads, -1).transpose(0, 2, 1, 3)
        values = values.reshape(B, L, self.n_kv_heads, -1).transpose(0, 2, 1, 3)

        if cache is not None:
            queries = self.rope(queries, offset=cache.offset)
            keys = self.rope(keys, offset=cache.offset)
            keys, values = cache.update_and_fetch(keys, values)
        else:
            queries = self.rope(queries)
            keys = self.rope(keys)

        if self.model_type == "gemma":
            output = scaled_dot_product_attention(
                queries, keys, values, cache, scale=self.scale, mask=mask
            )
        else:
            queries = queries * self.scale

            if self.repeats > 1:
                keys = mx.repeat(keys, self.repeats, axis=1)
                values = mx.repeat(values, self.repeats, axis=1)

            # Match HF eager semantics: compute logits in fp32, softmax in fp32,
            # then cast attention weights back to query dtype before value matmul.
            scores = queries.astype(mx.float32) @ keys.astype(mx.float32).swapaxes(
                -1, -2
            )
            if self.attn_logit_softcapping is not None:
                scores = mx.tanh(scores / self.attn_logit_softcapping)
                scores *= self.attn_logit_softcapping

            if mask is not None:
                # Match HF mask handling: accept attention-mask style [B, S] inputs,
                # precomputed additive masks, or 4D masks.
                if mask.ndim == 2:
                    if mask.shape[0] == B:
                        mask = mask[:, None, None, :]
                    else:
                        mask = mask[None, None, :, :]
                elif mask.ndim == 3:
                    mask = mask[:, None, :, :]

                mask = mask[..., : keys.shape[-2]]
                if mask.dtype not in (mx.float16, mx.bfloat16, mx.float32):
                    mask = mask.astype(scores.dtype)
                    mask = (1.0 - mask) * mx.finfo(scores.dtype).min
                else:
                    mask = mask.astype(scores.dtype)
                scores = scores + mask
            scores = mx.softmax(scores, precise=True, axis=-1).astype(queries.dtype)
            output = scores @ values

        output = output.transpose(0, 2, 1, 3).reshape(B, L, -1)
        return self.o_proj(output)


class MLP(nn.Module):
    def __init__(self, config: TextConfig):
        super().__init__()
        self.gate_proj = nn.Linear(
            config.hidden_size, config.intermediate_size, bias=False
        )
        self.down_proj = nn.Linear(
            config.intermediate_size, config.hidden_size, bias=False
        )
        self.up_proj = nn.Linear(
            config.hidden_size, config.intermediate_size, bias=False
        )
        self.activation = _activation_from_config(config)

    def __call__(self, x) -> mx.array:
        return self.down_proj(self.activation(self.gate_proj(x)) * self.up_proj(x))


class TransformerBlock(nn.Module):
    def __init__(self, config: TextConfig):
        super().__init__()
        self.model_type = config.model_type
        self.num_attention_heads = config.num_attention_heads
        self.hidden_size = config.hidden_size
        self.self_attn = Attention(config)
        self.mlp = MLP(config)
        self.input_layernorm = RMSNorm(config.hidden_size, eps=config.rms_norm_eps)
        self.post_attention_layernorm = RMSNorm(
            config.hidden_size, eps=config.rms_norm_eps
        )
        self.config = config

        if config.model_type == "gemma2":
            self.pre_feedforward_layernorm = RMSNorm(
                config.hidden_size, eps=config.rms_norm_eps
            )
            self.post_feedforward_layernorm = RMSNorm(
                config.hidden_size, eps=config.rms_norm_eps
            )

    def __call__(
        self,
        x: mx.array,
        mask: Optional[mx.array] = None,
        cache: Optional[KVCache] = None,
    ) -> mx.array:
        # Self attention block
        r = self.self_attn(self.input_layernorm(x), mask, cache)

        if self.model_type == "gemma":
            # Gemma: Post-attention residual connection then MLP
            h = x + r
            r = self.mlp(self.post_attention_layernorm(h))
            out = h + r
        else:
            # Gemma2: Normalized residual connections with pre/post norms
            h = x + self.post_attention_layernorm(r)
            r = self.mlp(self.pre_feedforward_layernorm(h))
            out = h + self.post_feedforward_layernorm(r)
        return out


class GemmaModel(nn.Module):
    def __init__(self, config: TextConfig):
        super().__init__()
        self.config = config
        self.vocab_size = config.vocab_size
        self.num_hidden_layers = config.num_hidden_layers
        assert self.vocab_size > 0
        self.embed_tokens = nn.Embedding(config.vocab_size, config.hidden_size)
        self.layers = [
            TransformerBlock(config=config) for _ in range(config.num_hidden_layers)
        ]
        self.norm = RMSNorm(config.hidden_size, eps=config.rms_norm_eps)

    def __call__(
        self,
        inputs: mx.array,
        inputs_embeds: Optional[mx.array] = None,
        mask: Optional[mx.array] = None,
        cache=None,
    ):
        # for passing merged input embeddings
        if inputs_embeds is None:
            h = self.embed_tokens(inputs)
        else:
            h = inputs_embeds

        normalizer = mx.array(self.config.hidden_size**0.5, dtype=h.dtype)
        h = h * normalizer

        if cache is None:
            cache = [None] * len(self.layers)

        use_bidirectional = bool(
            getattr(self.config, "use_bidirectional_attention", False)
        )
        if mask is None:
            if not use_bidirectional:
                mask = create_attention_mask(h, cache[0], return_array=True)
        elif not use_bidirectional and cache[0] is not None and cache[0].offset > 0:
            # For causal decoding, rebuild the mask against current cache length.
            mask = create_attention_mask(h, cache[0], return_array=True)

        for layer, c in zip(self.layers, cache):
            h = layer(h, mask, c)

        return self.norm(h)


class LanguageModel(nn.Module):
    def __init__(self, config: TextConfig):
        super().__init__()
        self.config = config
        self.final_logit_softcapping = config.final_logit_softcapping
        self.model_type = config.model_type
        self.model = GemmaModel(config)

        if self.model_type not in ["gemma", "gemma2"]:
            raise ValueError(
                f"Model type {self.model_type} not supported. Currently only 'gemma' is supported"
            )

    def __call__(
        self,
        inputs: mx.array,
        inputs_embeds: Optional[mx.array] = None,
        mask: Optional[mx.array] = None,
        cache=None,
        **kwargs,
    ):
        out = self.model(inputs, mask=mask, cache=cache, inputs_embeds=inputs_embeds)
        out = self.model.embed_tokens.as_linear(out)

        if self.model_type == "gemma2":
            out = mx.tanh(out / self.final_logit_softcapping)
            out = out * self.final_logit_softcapping
        return LanguageModelOutput(logits=out)

    def sanitize(self, weights):
        return {
            k: v for k, v in weights.items() if "self_attn.rotary_emb.inv_freq" not in k
        }

    @property
    def layers(self):
        return self.model.layers

    @property
    def head_dim(self):
        return (
            self.config.hidden_size // self.config.num_attention_heads
            if self.model_type == "gemma"
            else self.config.head_dim
        )

    @property
    def n_kv_heads(self):
        return self.config.num_key_value_heads



"""Vision encoder for Jina VLM in MLX."""

from typing import List, Tuple

import mlx.core as mx
import mlx.nn as nn

from .config import VisionConfig


class PatchEmbedding(nn.Module):
    """Patch embedding using linear projection."""

    def __init__(self, config: VisionConfig):
        super().__init__()
        self.patch_size = config.patch_size
        self.num_channels = config.num_channels
        self.hidden_size = config.hidden_size

        # Linear projection for patches - named to match weights
        patch_dim = config.num_channels * config.patch_size * config.patch_size
        self.proj = nn.Linear(patch_dim, config.hidden_size, bias=config.use_bias)

    def __call__(self, x: mx.array) -> Tuple[mx.array, Tuple[int, int]]:
        if x.ndim == 3:
            # Already patchified: (B, n_patches, patch_dim)
            B, n_patches, _ = x.shape
            nH = nW = int(n_patches**0.5)
            x = self.proj(x)
        else:
            # Image format: (B, C, H, W)
            B, C, H, W = x.shape
            pH, pW = self.patch_size, self.patch_size
            nH, nW = H // pH, W // pW
            x = x.reshape(B, C, nH, pH, nW, pW)
            x = x.transpose(0, 2, 4, 1, 3, 5)
            x = x.reshape(B, nH * nW, C * pH * pW)
            x = self.proj(x)
        return x, (nH, nW)


class VisionMLP(nn.Module):
    """MLP for vision transformer - matches weight naming: ffn.up, ffn.down"""

    def __init__(self, config: VisionConfig):
        super().__init__()
        # Named to match weights: ffn.up, ffn.down
        self.up = nn.Linear(
            config.hidden_size, config.intermediate_size, bias=config.use_bias
        )
        self.down = nn.Linear(
            config.intermediate_size, config.hidden_size, bias=config.use_bias
        )
        # Use built-in GELU with tanh approximation
        if config.activation == "gelu_pytorch_tanh":
            self.gelu = nn.GELU(approx="tanh")
        else:
            self.gelu = nn.GELU()

    def __call__(self, x: mx.array) -> mx.array:
        x = self.up(x)
        x = self.gelu(x)
        x = self.down(x)
        return x


class VisionAttention(nn.Module):
    """Multi-head self-attention - matches weight naming: attn.qkv, attn.out"""

    def __init__(self, config: VisionConfig):
        super().__init__()
        self.num_heads = config.num_attention_heads
        self.head_dim = config.head_dim
        self.scale = self.head_dim**-0.5

        # Fused QKV projection - named to match weights
        self.qkv = nn.Linear(
            config.hidden_size,
            3 * config.num_attention_heads * config.head_dim,
            bias=config.use_bias,
        )
        self.out = nn.Linear(
            config.num_attention_heads * config.head_dim,
            config.hidden_size,
            bias=config.use_bias,
        )

    def __call__(self, x: mx.array) -> mx.array:
        B, L, _ = x.shape
        qkv = self.qkv(x)
        qkv = qkv.reshape(B, L, 3, self.num_heads, self.head_dim)
        qkv = qkv.transpose(2, 0, 3, 1, 4)
        q, k, v = qkv[0], qkv[1], qkv[2]

        attn = (q @ k.transpose(0, 1, 3, 2)) * self.scale
        attn = mx.softmax(attn, axis=-1)
        x = attn @ v

        x = x.transpose(0, 2, 1, 3).reshape(B, L, -1)
        x = self.out(x)
        return x


class VisionEncoderLayer(nn.Module):
    """Transformer block - matches weight naming: attn_norm, ffn_norm"""

    def __init__(self, config: VisionConfig):
        super().__init__()
        # Named to match weights: attn_norm, ffn_norm
        self.attn_norm = nn.LayerNorm(
            config.hidden_size, eps=config.layer_norm_eps, bias=config.use_bias
        )
        self.attn = VisionAttention(config)
        self.ffn_norm = nn.LayerNorm(
            config.hidden_size, eps=config.layer_norm_eps, bias=config.use_bias
        )
        self.ffn = VisionMLP(config)

    def __call__(self, x: mx.array) -> mx.array:
        x = x + self.attn(self.attn_norm(x))
        x = x + self.ffn(self.ffn_norm(x))
        return x


class VisionModel(nn.Module):
    """Vision encoder (SigLIP-style ViT)."""

    def __init__(self, config: VisionConfig):
        super().__init__()
        self.config = config
        self.model_type = config.model_type
        self.hidden_size = config.hidden_size
        self.vit_layers = config.vit_layers

        # Named to match weights: patch_embed.proj
        self.patch_embed = PatchEmbedding(config)

        # Named to match weights: pos_embed (saved as 2D, not 3D)
        num_patches = (config.image_size // config.patch_size) ** 2
        if config.use_cls_token:
            num_patches += 1
            self.cls_token = mx.zeros((1, 1, config.hidden_size))
        else:
            self.cls_token = None
        self.pos_embed = mx.zeros((num_patches, config.hidden_size))

        # Transformer blocks
        self.layers = [
            VisionEncoderLayer(config) for _ in range(config.num_hidden_layers)
        ]

        # Named to match weights: post_norm
        if config.post_layer_norm:
            self.post_norm = nn.LayerNorm(
                config.hidden_size, eps=config.layer_norm_eps, bias=config.use_bias
            )
        else:
            self.post_norm = None

    def __call__(self, x: mx.array) -> Tuple[mx.array, List[mx.array]]:
        x, shape = self.patch_embed(x)

        if self.cls_token is not None:
            B = x.shape[0]
            cls = mx.broadcast_to(self.cls_token, (B, 1, self.hidden_size))
            x = mx.concatenate([cls, x], axis=1)

        # pos_embed is (num_patches, hidden_size), add batch dim for broadcast
        x = x + self.pos_embed[None, :, :]

        hidden_states = []
        for layer in self.layers:
            x = layer(x)
            hidden_states.append(x)

        if self.post_norm is not None:
            x = self.post_norm(x)
            hidden_states.append(x)

        return x, hidden_states

    def get_features(self, images: mx.array) -> mx.array:
        """Extract features from specific ViT layers.

        Note: hidden_states includes all layer outputs plus the post_norm output.
        vit_layers indices (e.g., [-4, -10]) are applied to this full list.
        For 27 layers with post_norm, hidden_states has 28 elements:
        - indices 0-26: layer 0-26 outputs
        - index 27: post_norm output
        So vit_layers=[-4, -10] extracts layers 24 and 18 (not 23 and 17).
        """
        _, hidden_states = self(images)
        # Use full hidden_states including post_norm output for correct indexing

        features = []
        for layer_idx in self.vit_layers:
            feats = hidden_states[layer_idx]
            if self.cls_token is not None:
                feats = feats[:, 1:]
            features.append(feats)

        return mx.concatenate(features, axis=-1)



import re
from typing import Optional

import mlx.core as mx
import mlx.nn as nn
import numpy as np

from ..base import InputEmbeddingsFeatures
from .config import ModelConfig
from .language import LanguageModel
from .vision import CallableModuleList, VisionModel


def build_vision_projector(config):
    hidden_size = config.text_config.hidden_size
    projector_type = getattr(config, "mm_projector_type", "mlp2x_gelu")
    if projector_type == "linear":
        return nn.Linear(config.mm_hidden_size, hidden_size)

    mlp_gelu_match = re.match(r"^mlp(\d+)x_gelu$", projector_type)
    if mlp_gelu_match:
        mlp_depth = int(mlp_gelu_match.group(1))
        modules = CallableModuleList()
        modules.append(nn.Linear(config.mm_hidden_size, hidden_size))
        for _ in range(1, mlp_depth):
            modules.append(nn.GELU())
            modules.append(nn.Linear(hidden_size, hidden_size))
        return modules
    raise ValueError(f"Unknown projector type: {projector_type}")


class Model(nn.Module):
    def __init__(self, config: ModelConfig):
        super().__init__()
        self.config = config
        self.vision_tower = VisionModel(config.vision_config)
        self.language_model = LanguageModel(config.text_config)
        self.mm_projector = build_vision_projector(config)

    def get_input_embeddings(
        self,
        input_ids: Optional[mx.array] = None,
        pixel_values: Optional[mx.array] = None,
        mask: Optional[mx.array] = None,
        **kwargs,
    ):
        if pixel_values is None:
            return InputEmbeddingsFeatures(
                inputs_embeds=self.language_model.model.embed_tokens(input_ids)
            )

        # Cast pixel_values to model dtype for consistent dtype propagation
        model_dtype = self.language_model.model.embed_tokens.weight.dtype
        pixel_values = pixel_values.astype(model_dtype)

        _, image_features, _ = self.vision_tower(pixel_values.transpose(0, 2, 3, 1))
        B, H, W, C = image_features.shape
        image_features = image_features.reshape(B, H * W, C)
        image_features = self.mm_projector(image_features)

        final_inputs_embeds = self.prepare_inputs_for_multimodal(
            image_features, input_ids, mask
        )
        return InputEmbeddingsFeatures(inputs_embeds=final_inputs_embeds)

    # Source: https://github.com/apple/ml-fastvlm/blob/592b4add3c1c8a518e77d95dc6248e76c1dd591f/llava/model/llava_arch.py#L146
    def prepare_inputs_for_multimodal(self, image_features, input_ids, mask):
        # Batch evaluation to avoid repeated sync points
        if mask is not None:
            starts = mx.argmax(mask, axis=-1)
            lengths = mask.sum(axis=-1)
            mx.eval(starts, lengths)
            starts = starts.tolist()
            lengths = lengths.tolist()
            input_ids = [
                cur_input_ids[start : start + length]
                for cur_input_ids, start, length in zip(input_ids, starts, lengths)
            ]

        # Pre-compute image token masks and counts for all batch items (single eval)
        image_token_index = self.config.image_token_index
        image_masks = [
            cur_input_ids == image_token_index for cur_input_ids in input_ids
        ]
        num_images_list = [m.sum() for m in image_masks]
        mx.eval(image_masks + num_images_list)
        num_images_list = [n.item() for n in num_images_list]

        new_input_embeds = []
        cur_image_idx = 0
        for batch_idx, cur_input_ids in enumerate(input_ids):
            num_images = num_images_list[batch_idx]
            if num_images == 0:
                cur_image_features = image_features[cur_image_idx]
                cur_input_embeds_1 = self.language_model.model.embed_tokens(
                    cur_input_ids
                )
                cur_input_embeds = mx.concatenate(
                    [cur_input_embeds_1, cur_image_features[0:0]], axis=0
                )
                new_input_embeds.append(cur_input_embeds)
                cur_image_idx += 1
                continue

            # Find image token indices - mask was already evaluated above
            image_token_mask = image_masks[batch_idx]
            image_token_indices = (
                [-1]
                + np.where(np.array(image_token_mask))[0].tolist()
                + [cur_input_ids.shape[0]]
            )

            cur_input_ids_noim = []
            for i in range(len(image_token_indices) - 1):
                cur_input_ids_noim.append(
                    cur_input_ids[
                        image_token_indices[i] + 1 : image_token_indices[i + 1]
                    ]
                )
            split_sizes = image_token_indices[1:]
            cur_input_embeds = self.language_model.model.embed_tokens(
                mx.concatenate(cur_input_ids_noim)
            )
            cur_input_embeds_no_im = mx.split(cur_input_embeds, split_sizes)

            cur_new_input_embeds = []
            for i in range(num_images + 1):
                cur_new_input_embeds.append(cur_input_embeds_no_im[i])
                if i < num_images:
                    cur_image_features = image_features[cur_image_idx]
                    cur_image_idx += 1
                    cur_new_input_embeds.append(cur_image_features)
            cur_new_input_embeds = mx.concatenate(cur_new_input_embeds)

            new_input_embeds.append(cur_new_input_embeds)

        if self.config.tokenizer_model_max_length is not None:
            new_input_embeds = [
                x[: self.config.tokenizer_model_max_length] for x in new_input_embeds
            ]

        max_len = max(x.shape[0] for x in new_input_embeds)
        new_input_embeds_padded = []
        for i, cur_new_embed in enumerate(new_input_embeds):
            cur_len = cur_new_embed.shape[0]
            padded = cur_new_embed
            if max_len > cur_len:
                if self.config.tokenizer_padding_side == "left":
                    padded = mx.concatenate(
                        (
                            mx.zeros(
                                (max_len - cur_len, cur_new_embed.shape[1]),
                                dtype=cur_new_embed.dtype,
                            ),
                            cur_new_embed,
                        ),
                        axis=0,
                    )
                else:
                    padded = mx.concatenate(
                        (
                            cur_new_embed,
                            mx.zeros(
                                (max_len - cur_len, cur_new_embed.shape[1]),
                                dtype=cur_new_embed.dtype,
                            ),
                        ),
                        axis=0,
                    )
            new_input_embeds_padded.append(padded)
        new_input_embeds = mx.stack(new_input_embeds_padded)
        return new_input_embeds

    @property
    def layers(self):
        return self.language_model.model.layers

    def __call__(
        self,
        input_ids: mx.array,
        pixel_values: mx.array,
        mask: mx.array,
        cache=None,
        **kwargs,
    ):
        input_embeddings_features = self.get_input_embeddings(
            input_ids, pixel_values, mask
        )
        logits = self.language_model(
            input_ids,
            mask=mask,
            cache=cache,
            inputs_embeds=input_embeddings_features.inputs_embeds,
        )
        return logits

    def sanitize(self, weights):
        def transform_key(key):
            if "vision_tower" in key:
                if "model.vision_tower" in key:
                    key = key.replace(
                        "model.vision_tower.vision_tower.model",
                        "vision_tower.vision_model",
                    )
                    key = key.replace("patch_embed", "patch_embed.blocks")
                return key
            if "lm_head" in key:
                return key
            if "mm_projector" in key:
                return key.replace("model.", "")
            if "language_model" not in key:
                return "language_model." + key
            return key

        return {transform_key(k): v for k, v in weights.items()}



import re
from functools import partial, reduce
from typing import Optional, Tuple

import mlx.core as mx
import mlx.nn as nn
from PIL import Image
from transformers.image_transforms import (
    convert_to_rgb,
    normalize,
    rescale,
    resize,
    to_channel_dimension_format,
)
from transformers.image_utils import to_numpy_array

from ..base import BaseImageProcessor, InputEmbeddingsFeatures
from .config import ModelConfig, VisionConfig
from .language import LanguageModel
from .vision import VisionModel


class ImageProcessor(BaseImageProcessor):
    def preprocess(self, images):
        if isinstance(images, Image.Image):
            images = [images]
        else:
            assert isinstance(images, list)

        transforms = [
            convert_to_rgb,
            to_numpy_array,
            partial(
                resize,
                size=self.size,
                resample=self.resample,
                data_format=self.data_format,
            ),
            partial(rescale, scale=self.rescale_factor, data_format=self.data_format),
            partial(
                normalize,
                mean=self.image_mean,
                std=self.image_std,
                data_format=self.data_format,
            ),
            partial(
                to_channel_dimension_format,
                channel_dim=self.data_format,
                input_channel_dim=self.data_format,
            ),
        ]

        images = reduce(lambda x, f: [*map(f, x)], transforms, images)

        return images


class LlavaMultiModalProjector(nn.Module):
    def __init__(self, config: ModelConfig):
        super().__init__()
        self.linear_1 = nn.Linear(
            config.vision_config.hidden_size, config.text_config.hidden_size, bias=True
        )
        self.gelu = nn.GELU()
        self.linear_2 = nn.Linear(
            config.text_config.hidden_size, config.text_config.hidden_size, bias=True
        )

    def __call__(self, x: mx.array) -> mx.array:
        x = self.linear_1(x)
        x = self.gelu(x)
        x = self.linear_2(x)
        return x


class SigLipVisionTower(nn.Module):
    def __init__(self, config: VisionConfig):
        super().__init__()
        self.vision_tower = VisionModel(config)

    def __call__(
        self, x: mx.array, output_hidden_states: Optional[bool] = None
    ) -> mx.array:
        return self.vision_tower(x, output_hidden_states)


class Model(nn.Module):
    def __init__(self, config: ModelConfig):
        super().__init__()
        self.model_type = config.model_type
        self.config = config

        self.vision_tower = SigLipVisionTower(config.vision_config)
        self.language_model = LanguageModel(config.text_config)
        self.mm_projector = LlavaMultiModalProjector(config)

    def get_input_embeddings(
        self,
        input_ids: Optional[mx.array] = None,
        pixel_values: Optional[mx.array] = None,
        **kwargs,
    ):
        if pixel_values is None:
            return InputEmbeddingsFeatures(
                inputs_embeds=self.language_model.model.embed_tokens(input_ids)
            )

        inputs_embeds = self.language_model.model.embed_tokens(input_ids)

        *_, hidden_state = self.vision_tower(
            pixel_values.transpose(0, 2, 3, 1), output_hidden_states=True
        )

        image_features = hidden_state[-1].astype(pixel_values.dtype)
        assert image_features.shape[-2] == 729

        image_features = self.mm_projector(image_features)

        final_inputs_embeds = self._prepare_inputs_for_multimodal(
            image_features, inputs_embeds, input_ids
        )
        return InputEmbeddingsFeatures(inputs_embeds=final_inputs_embeds)

    def _prepare_inputs_for_multimodal(self, image_features, inputs_embeds, input_ids):
        image_token_index = self.config.image_token_index
        num_images, num_image_patches, embed_dim = image_features.shape

        batch_size, seq_length, embed_dim = inputs_embeds.shape
        num_images, num_image_patches, _ = image_features.shape

        # Positions of <image> tokens in input_ids for each batch
        image_positions = mx.argmax(input_ids == image_token_index, axis=1)

        final_embeddings = []
        for b in range(batch_size):
            text_segments = []
            start_idx = 0
            position = int(image_positions[b].item())

            text_segments.append(inputs_embeds[b : b + 1, start_idx:position])
            text_segments.append(image_features[b : b + 1])
            text_segments.append(inputs_embeds[b : b + 1, position + 1 :])

            batch_embeddings = mx.concatenate(text_segments, axis=1)
            final_embeddings.append(batch_embeddings)

        # Create a final embedding of shape
        # (batch_size, num_image_patches + sequence_len, embed_dim)
        return mx.concatenate(final_embeddings, axis=0)

    @property
    def layers(self):
        return self.language_model.model.layers

    def __call__(
        self,
        input_ids: mx.array,
        pixel_values: mx.array,
        mask: Optional[mx.array] = None,
        cache: Optional[Tuple[mx.array, mx.array]] = None,
        **kwargs,
    ):
        input_embeddings_features = self.get_input_embeddings(input_ids, pixel_values)
        logits = self.language_model(
            inputs=input_ids,
            cache=cache,
            inputs_embeds=input_embeddings_features.inputs_embeds,
            mask=None,  # TODO: add mask
        )
        return logits

    def sanitize(self, weights):
        weights = {
            (
                f"{k.split('.', 1)[1]}"
                if re.match(r"^model\.vision_tower", k)
                else (
                    f"mm_projector.linear_1.{k.split('.')[-1]}"
                    if re.match(r"^model\.mm_projector\.0", k)
                    else (
                        f"mm_projector.linear_2.{k.split('.')[-1]}"
                        if re.match(r"^model\.mm_projector\.2", k)
                        else (
                            f"language_model.model.{k}"
                            if re.match(r"^lm_head", k)
                            else (
                                f"language_model.{k}"
                                if re.match(r"^model\.(embed_tokens|norm|layers)", k)
                                else k
                            )
                        )
                    )
                )
            ): v
            for k, v in weights.items()
        }

        weights = {
            (
                f"vision_tower.vision_tower.vision_model.head.attention.in_proj.bias"
                if re.match(
                    r"^vision_tower\.vision_tower\.vision_model\.head\.attention\.in_proj_bias",
                    k,
                )
                else (
                    f"vision_tower.vision_tower.vision_model.head.attention.in_proj.weight"
                    if re.match(
                        r"^vision_tower\.vision_tower\.vision_model\.head\.attention\.in_proj_weight",
                        k,
                    )
                    else k
                )
            ): v
            for k, v in weights.items()
        }

        return weights



"""Language model decoder for Jina VLM in MLX."""

from typing import List, Optional

import mlx.core as mx
import mlx.nn as nn
from mlx_lm.models.base import create_attention_mask, scaled_dot_product_attention
from mlx_lm.models.cache import KVCache

from ..base import LanguageModelOutput
from .config import TextConfig


class RMSNorm(nn.Module):
    """RMS Layer Normalization."""

    def __init__(self, dims: int, eps: float = 1e-6):
        super().__init__()
        self.eps = eps
        self.weight = mx.ones((dims,))

    def __call__(self, x: mx.array) -> mx.array:
        rms = mx.sqrt(mx.mean(x * x, axis=-1, keepdims=True) + self.eps)
        return self.weight * (x / rms)


class RoPE(nn.Module):
    """Rotary Positional Embeddings."""

    def __init__(self, dims: int, theta: float = 1000000.0):
        super().__init__()
        self.dims = dims
        self.theta = theta
        inv_freq = 1.0 / (theta ** (mx.arange(0, dims, 2).astype(mx.float32) / dims))
        self._inv_freq = inv_freq

    def __call__(self, x: mx.array, offset: int = 0) -> mx.array:
        seq_len = x.shape[2]
        positions = mx.arange(offset, offset + seq_len).astype(mx.float32)
        freqs = positions[:, None] * self._inv_freq[None, :]
        emb = mx.concatenate([freqs, freqs], axis=-1)
        cos = mx.cos(emb)[None, None, :, :]
        sin = mx.sin(emb)[None, None, :, :]
        x1 = x[..., : self.dims // 2]
        x2 = x[..., self.dims // 2 :]
        rotated = mx.concatenate([-x2, x1], axis=-1)
        return (x * cos + rotated * sin).astype(x.dtype)


class Attention(nn.Module):
    """Multi-head attention with GQA and RoPE - matches weight naming: attn.qkv, attn.out"""

    def __init__(self, config: TextConfig):
        super().__init__()
        self.config = config
        self.num_heads = config.num_attention_heads
        self.num_kv_heads = config.num_key_value_heads
        self.head_dim = config.head_dim
        self.num_kv_groups = self.num_heads // self.num_kv_heads
        self.scale = self.head_dim**-0.5

        # Fused QKV projection - named to match weights
        qkv_size = (
            config.num_attention_heads + 2 * config.num_key_value_heads
        ) * config.head_dim
        self.qkv = nn.Linear(config.hidden_size, qkv_size, bias=False)
        self.out = nn.Linear(
            config.num_attention_heads * config.head_dim, config.hidden_size, bias=False
        )

        # QK normalization - named to match weights
        if config.use_qk_norm:
            self.q_norm = RMSNorm(config.head_dim, eps=config.rms_norm_eps)
            self.k_norm = RMSNorm(config.head_dim, eps=config.rms_norm_eps)
        else:
            self.q_norm = None
            self.k_norm = None

        self.rope = RoPE(config.head_dim, theta=config.rope_theta)

    def __call__(
        self,
        x: mx.array,
        mask: Optional[mx.array] = None,
        cache: Optional[KVCache] = None,
    ) -> mx.array:
        B, L, _ = x.shape

        # Compute fused QKV
        qkv = self.qkv(x)
        q_size = self.num_heads * self.head_dim
        kv_size = self.num_kv_heads * self.head_dim

        q = qkv[..., :q_size]
        k = qkv[..., q_size : q_size + kv_size]
        v = qkv[..., q_size + kv_size :]

        q = q.reshape(B, L, self.num_heads, self.head_dim).transpose(0, 2, 1, 3)
        k = k.reshape(B, L, self.num_kv_heads, self.head_dim).transpose(0, 2, 1, 3)
        v = v.reshape(B, L, self.num_kv_heads, self.head_dim).transpose(0, 2, 1, 3)

        if self.q_norm is not None:
            q = self.q_norm(q)
            k = self.k_norm(k)

        if cache is not None:
            q = self.rope(q, offset=cache.offset)
            k = self.rope(k, offset=cache.offset)
            k, v = cache.update_and_fetch(k, v)
        else:
            q = self.rope(q)
            k = self.rope(k)

        output = scaled_dot_product_attention(
            q, k, v, cache=cache, scale=self.scale, mask=mask
        )
        output = output.transpose(0, 2, 1, 3).reshape(B, L, -1)
        return self.out(output)


class MLP(nn.Module):
    """MLP with SwiGLU - matches weight naming: ffn.gate_up, ffn.down"""

    def __init__(self, config: TextConfig):
        super().__init__()
        # Fused gate and up projection - named to match weights
        self.gate_up = nn.Linear(
            config.hidden_size, 2 * config.intermediate_size, bias=False
        )
        self.down = nn.Linear(config.intermediate_size, config.hidden_size, bias=False)

    def __call__(self, x: mx.array) -> mx.array:
        gate_up = self.gate_up(x)
        # Jina VLM convention: first half is value, second half is gate (activated)
        up, gate = mx.split(gate_up, 2, axis=-1)
        return self.down(nn.silu(gate) * up)


class TransformerBlock(nn.Module):
    """Transformer block - matches weight naming: attn_norm, ffn_norm"""

    def __init__(self, config: TextConfig, layer_idx: int = 0):
        super().__init__()
        # Named to match weights
        self.attn_norm = RMSNorm(config.hidden_size, eps=config.rms_norm_eps)
        self.attn = Attention(config)
        self.ffn_norm = RMSNorm(config.hidden_size, eps=config.rms_norm_eps)
        self.ffn = MLP(config)

    def __call__(
        self,
        x: mx.array,
        mask: Optional[mx.array] = None,
        cache: Optional[KVCache] = None,
    ) -> mx.array:
        h = self.attn(self.attn_norm(x), mask=mask, cache=cache)
        x = x + h
        x = x + self.ffn(self.ffn_norm(x))
        return x


class ExtendedEmbedding(nn.Module):
    """Embedding with additional tokens - matches weight naming: embedding, new_embedding"""

    def __init__(self, vocab_size: int, additional_size: int, dims: int):
        super().__init__()
        self.vocab_size = vocab_size
        self.additional_size = additional_size
        # Named to match weights
        self.embedding = mx.zeros((vocab_size, dims))
        self.new_embedding = mx.zeros((additional_size, dims))

    def __call__(self, x: mx.array) -> mx.array:
        full_embedding = mx.concatenate([self.embedding, self.new_embedding], axis=0)
        return full_embedding[x]


class TextModel(nn.Module):
    """Text decoder model - matches weight naming structure"""

    def __init__(self, config: TextConfig):
        super().__init__()
        self.config = config

        # Named to match weights: language_model.embedding
        if config.additional_vocab_size > 0:
            self.embedding = ExtendedEmbedding(
                config.vocab_size, config.additional_vocab_size, config.hidden_size
            )
        else:
            self.embedding = nn.Embedding(config.vocab_size, config.hidden_size)

        self.layers = [
            TransformerBlock(config, layer_idx=i)
            for i in range(config.num_hidden_layers)
        ]

        # Named to match weights: language_model.ln_f
        self.ln_f = RMSNorm(config.hidden_size, eps=config.rms_norm_eps)

    def __call__(
        self,
        input_ids: mx.array,
        inputs_embeds: Optional[mx.array] = None,
        mask: Optional[mx.array] = None,
        cache: Optional[List[KVCache]] = None,
    ) -> mx.array:
        if inputs_embeds is None:
            x = self.embedding(input_ids)
        else:
            x = inputs_embeds

        for i, layer in enumerate(self.layers):
            layer_cache = cache[i] if cache is not None else None
            x = layer(x, mask=mask, cache=layer_cache)

        return self.ln_f(x)


class LanguageModel(nn.Module):
    """Language model wrapper - the TextModel is accessed as language_model in weights"""

    def __init__(self, config: TextConfig):
        super().__init__()
        self.config = config
        self.model_type = config.model_type
        # This will be loaded under "language_model" prefix
        self.embedding = None  # Handled by sanitize
        self.layers = None  # Handled by sanitize
        self.ln_f = None  # Handled by sanitize

        # Build the actual model components directly here
        # They'll be found via language_model.embedding, language_model.layers, etc.
        if config.additional_vocab_size > 0:
            self.embedding = ExtendedEmbedding(
                config.vocab_size, config.additional_vocab_size, config.hidden_size
            )
        else:
            self.embedding = nn.Embedding(config.vocab_size, config.hidden_size)

        self.layers = [
            TransformerBlock(config, layer_idx=i)
            for i in range(config.num_hidden_layers)
        ]
        self.ln_f = RMSNorm(config.hidden_size, eps=config.rms_norm_eps)

    def __call__(
        self,
        inputs: mx.array,
        inputs_embeds: Optional[mx.array] = None,
        mask: Optional[mx.array] = None,
        cache: Optional[List[KVCache]] = None,
        **kwargs,
    ) -> LanguageModelOutput:
        if inputs_embeds is None:
            x = self.embedding(inputs)
        else:
            x = inputs_embeds

        # Initialize cache if needed
        if cache is None:
            cache = [None] * len(self.layers)

        # Create causal attention mask
        mask = create_attention_mask(x, cache)

        for i, layer in enumerate(self.layers):
            x = layer(x, mask=mask, cache=cache[i])

        hidden_states = self.ln_f(x)
        logits = self.lm_head(hidden_states)
        return LanguageModelOutput(logits=logits)



import argparse
import csv
import json
import logging
import random
import traceback
from pathlib import Path
from typing import Optional

import mlx.core as mx
from datasets import load_dataset
from tqdm import tqdm

from mlx_vlm import load
from mlx_vlm.evals.utils import inference
from mlx_vlm.generate import batch_generate
from mlx_vlm.sample_utils import top_p_sampling


def process_question(sample: dict) -> str:
    """Format the question."""
    return sample["question"]


def normalize_answer(response: str, problem: dict) -> Optional[str]:
    """Normalize the model's response to extract the answer."""
    if not response:
        return None
    return response.strip()


def evaluate_answer(prediction: Optional[str], ground_truth: list) -> bool:
    """Check if any ground truth answer is contained in the prediction."""
    if prediction is None:
        return False
    pred = prediction.strip().lower()
    return any(str(a).strip().lower() in pred for a in ground_truth)


def OCRBench_val(results_list, args, model_name, dataset="OCRBench"):
    correct = 0
    total = len(results_list)
    category_scores = {}
    for row in results_list:
        ground_truth = row["ground_truth"]
        if isinstance(ground_truth, str):

            ground_truth = [a.strip() for a in ground_truth.split(";")]
        prediction = row["prediction"]

        is_correct = evaluate_answer(prediction, ground_truth)
        row["correct"] = is_correct
        if is_correct:
            correct += 1
        category = row["type"]
        if category not in category_scores:
            category_scores[category] = {"correct": 0, "total": 0}
        category_scores[category]["total"] += 1
        if is_correct:
            category_scores[category]["correct"] += 1

    accuracy = correct / total if total > 0 else 0

    output_dir = Path(args.output_dir)
    output_dir.mkdir(parents=True, exist_ok=True)

    results_file = output_dir / f"{model_name}_{dataset}_{args.split}.csv"

    fieldnames = [
        "id",
        "question",
        "dataset",
        "type",
        "ground_truth",
        "response",
        "prediction",
        "correct",
    ]

    with open(results_file, "w", newline="", encoding="utf-8") as f:
        writer = csv.DictWriter(f, fieldnames=fieldnames)
        writer.writeheader()
        for row in results_list:
            out_row = row.copy()
            if isinstance(out_row["ground_truth"], list):
                out_row["ground_truth"] = "; ".join(map(str, out_row["ground_truth"]))
            writer.writerow(out_row)

    summary = {
        "model": model_name,
        "dataset": args.dataset,
        "split": args.split,
        "total_samples": total,
        "correct": correct,
        "accuracy": accuracy,
        "category_scores": category_scores,
    }

    summary_file = output_dir / f"{model_name}_{dataset}_{args.split}.json"
    with open(summary_file, "w") as f:
        json.dump(summary, f, indent=2)

    print(f"\n{'='*80}")
    print(f"{dataset} Evaluation Results")
    print(f"{'='*80}")
    print(f"Model: {summary['model']}")
    print(f"Split: {args.split}")
    print(f"Total Samples: {total}")
    print(f"Correct: {correct}")
    print(f"Accuracy: {accuracy*100:.2f}%")

    if len(category_scores.items()) > 1:
        print("\n" + "-" * 80)
        print(f"Subcategory Scores:")
        print(f"{'-'*80}")
        for category, scores in category_scores.items():
            cat_total = scores["total"]
            cat_correct = scores["correct"]
            cat_accuracy = cat_correct / cat_total if cat_total > 0 else 0
            print(f"  {category}: {cat_correct}/{cat_total} ({cat_accuracy*100:.2f}%)")
        print(f"{'='*80}")
        print(f"\nResults saved to {results_file} and {summary_file}")


def parse_args():
    parser = argparse.ArgumentParser(
        description="Evaluate models on OCRBench benchmark"
    )
    parser.add_argument(
        "--model",
        type=str,
        required=True,
        help="The path to the MLX VLM model",
    )
    parser.add_argument(
        "--adapter-path",
        type=str,
        help="Optional path for the trained adapter weights and config",
    )
    parser.add_argument(
        "--dataset",
        type=str,
        default="echo840/OCRBench",
        help="Hugging Face dataset name",
    )
    parser.add_argument(
        "--split",
        type=str,
        default="test",
        choices=["test"],
        help="Dataset split to evaluate on",
    )
    parser.add_argument(
        "--streaming",
        action="store_true",
        help="Use streaming dataset loading",
    )
    parser.add_argument(
        "--max-samples",
        type=int,
        default=None,
        help="Maximum number of samples to evaluate (for debugging)",
    )
    parser.add_argument(
        "--predictions-file",
        type=str,
        default=None,
        help="File with predictions",
    )
    parser.add_argument(
        "--output-dir",
        type=str,
        default="results/ocrbench",
        help="Directory to save results",
    )
    parser.add_argument(
        "--max-tokens",
        type=int,
        default=512,
        help="Maximum number of tokens to generate",
    )
    parser.add_argument(
        "--temperature",
        type=float,
        default=0.0,
        help="Temperature for generation",
    )
    parser.add_argument(
        "--verbose",
        action="store_true",
        help="Print detailed output for debugging",
    )
    parser.add_argument("--seed", type=int, default=42, help="Random seed")
    parser.add_argument(
        "--batch-size",
        type=int,
        default=1,
        help="Batch size for generation (1 = sequential, >1 = batch generation)",
    )
    return parser.parse_args()


def create_sampler(temperature: float, top_p: float = 1.0):
    """Create a sampler function for batch generation.

    For accuracy consistency across batch sizes, we use deterministic sampling
    (temperature=0) by default. This ensures the same outputs regardless of batch size.
    """

    def sampler(logits: mx.array) -> mx.array:
        if temperature == 0:
            return mx.argmax(logits, axis=-1)
        else:
            if top_p > 0 and top_p < 1.0:
                return top_p_sampling(logits, top_p, temperature)
            else:
                return mx.random.categorical(logits * (1 / temperature))

    return sampler


def process_batch(
    model,
    processor,
    batch_samples,
    args,
):
    """Process a batch of samples using batch_generate.

    batch_generate now handles image size sorting internally to minimize
    padding effects and maintain accuracy.
    """
    prompts = []
    images = []
    sample_metadata = []

    for sample in batch_samples:
        pid = sample.get("id", str(sample.get("_idx", 0)))

        # Load and process image
        if "image" in sample and sample["image"]:
            image = sample["image"].convert("RGB")
        else:
            logging.warning(f"No image for sample {pid}, skipping")
            continue

        images.append(image)

        # Create prompt
        prompt = process_question(sample)
        prompts.append(prompt)

        # Store metadata for results
        sample_metadata.append(
            {
                "id": pid,
                "question": sample["question"],
                "dataset": sample.get("dataset", ""),
                "type": sample.get("type", ""),
                "ground_truth": (
                    sample.get("answers", [])
                    if hasattr(sample, "answers")
                    else sample.get("answer", [])
                ),
            }
        )

    if not prompts:
        return []

    # Create sampler for deterministic output (temperature=0 by default)
    sampler = create_sampler(args.temperature)

    # Use batch_generate for processing
    # batch_generate now handles image size sorting internally to avoid padding issues
    batch_response = batch_generate(
        model,
        processor,
        images=images,
        prompts=prompts,
        max_tokens=args.max_tokens,
        sampler=sampler,
        verbose=args.verbose,
    )

    # Process results
    results = []
    for text, metadata in zip(batch_response.texts, sample_metadata):
        response = text.strip()
        prediction = normalize_answer(response, {"question": metadata["question"]})

        result = {
            **metadata,
            "response": response,
            "prediction": prediction,
            "correct": False,
        }
        results.append(result)

        if args.verbose:
            logging.info(f"\nSample {metadata['id']}:")
            logging.info(f"Question: {metadata['question']}")
            logging.info(f"Response: {response}")
            logging.info(f"Prediction: {prediction}")
            logging.info(f"Ground Truth: {metadata['ground_truth']}")

    return results


def main():
    args = parse_args()

    random.seed(args.seed)

    # Setup logging
    logging.basicConfig(
        level=logging.INFO if args.verbose else logging.WARNING,
        format="%(asctime)s - %(levelname)s - %(message)s",
    )

    if args.predictions_file:
        logging.info(
            f"\033[32mLoading predictions from {args.predictions_file} for evaluation\033[0m"
        )
        with open(args.predictions_file, "r", encoding="utf-8") as f:
            reader = csv.DictReader(f)
            loaded_results = list(reader)
        model_name = Path(args.predictions_file).stem.split("_OCRBench")[0]
        dataset = (
            "OCRBench-v2" if "OCRBench-v2" in args.predictions_file else "OCRBench"
        )
        OCRBench_val(loaded_results, args, model_name, dataset)
        logging.info(f"\033[32mEvaluation complete\033[0m")
        return

    logging.info(f"Loading model from {args.model}")
    model, processor = load(
        args.model, adapter_path=args.adapter_path, trust_remote_code=True
    )

    # Load dataset
    logging.info(f"Loading dataset {args.dataset}, split {args.split}")
    dataset = load_dataset(args.dataset, split=args.split, streaming=args.streaming)

    if args.max_samples:
        dataset = dataset.take(args.max_samples)

    # Convert to list for batching if streaming
    if args.streaming:
        dataset = list(dataset)

    results = {}
    batch_size = args.batch_size

    if batch_size > 1:
        # Batch generation mode
        logging.info(f"Using batch generation with batch_size={batch_size}")

        # Collect samples into batches
        batch = []
        all_samples = list(dataset) if hasattr(dataset, "__iter__") else dataset

        # Add index to samples for tracking
        for idx, sample in enumerate(all_samples):
            sample["_idx"] = idx

        for idx, sample in enumerate(
            tqdm(all_samples, desc=f"Evaluating (batch_size={batch_size})")
        ):
            batch.append(sample)

            # Process batch when full or at the end
            if len(batch) >= batch_size or idx == len(all_samples) - 1:
                try:
                    batch_results = process_batch(model, processor, batch, args)
                    for result in batch_results:
                        results[result["id"]] = result
                except Exception as e:
                    logging.error(f"Error processing batch: {e}")
                    traceback.print_exc()

                batch = []

                # Clear memory after each batch
                mx.clear_cache()

    else:
        # Sequential generation mode (original behavior)
        for idx, sample in enumerate(tqdm(dataset, desc="Evaluating")):
            pid = sample.get("id", str(idx))

            try:
                # Load and process image
                if "image" in sample and sample["image"]:
                    image = sample["image"].convert("RGB")
                else:
                    logging.warning(f"No image for sample {pid}, skipping")
                    continue

                # Create prompt
                prompt = process_question(sample)

                # Generate response
                output = inference(
                    model,
                    processor,
                    prompt,
                    image=image,
                    max_tokens=args.max_tokens,
                    temperature=args.temperature,
                )

                response = output.strip()

                # Normalize answer
                prediction = normalize_answer(response, sample)

                # Store results (evaluation happens later)
                results[pid] = {
                    "id": pid,
                    "question": sample["question"],
                    "dataset": sample.get("dataset", ""),
                    "type": sample.get("type", ""),
                    "ground_truth": (
                        sample.get("answers", [])
                        if hasattr(sample, "answers")
                        else sample.get("answer", [])
                    ),
                    "response": response,
                    "prediction": prediction,
                    "correct": False,
                }

                if args.verbose:
                    logging.info(f"\nSample {pid}:")
                    logging.info(f"Question: {sample['question']}")
                    logging.info(f"Response: {response}")
                    logging.info(f"Prediction: {prediction}")
                    logging.info(f"Ground Truth: {sample.get('answers', [])}")

            except Exception as e:
                traceback.print_exc()
                logging.error(f"Error processing sample {pid}: {e}")
                continue

    results_list = list(results.values())
    model_name = args.model.split("/")[-1]
    dataset = args.dataset.split("/")[-1]
    OCRBench_val(results_list, args, model_name, dataset)


if __name__ == "__main__":
    main()



from .config import ModelConfig, TextConfig, VisionConfig
from .hunyuan_vl import LanguageModel, Model, VisionModel
from .processing_hunyuan_vl import (
    HunYuanVLImageProcessor,
    HunYuanVLProcessor,
    ImageProcessor,
)



"""Image processor for Jina VLM in MLX-VLM."""

import math
from typing import Dict, List, Optional, Tuple, Union

import mlx.core as mx
import numpy as np
from PIL import Image

# CLIP normalization constants
CLIP_MEAN = [0.48145466, 0.4578275, 0.40821073]
CLIP_STD = [0.26862954, 0.26130258, 0.27577711]

# Default special token IDs
DEFAULT_PATCH_TOKEN_ID = 151665  # <im_patch>
DEFAULT_START_TOKEN_ID = 151666  # <im_start>
DEFAULT_END_TOKEN_ID = 151667  # <im_end>
DEFAULT_COLUMN_TOKEN_ID = 151668  # <im_col>


def smart_resize(
    height: int,
    width: int,
    factor: int = 28,
    min_pixels: int = 56 * 56,
    max_pixels: int = 14 * 14 * 4 * 1280,
) -> Tuple[int, int]:
    """Resize dimensions while maintaining aspect ratio and constraints."""
    h_bar = round(height / factor) * factor
    w_bar = round(width / factor) * factor

    if h_bar * w_bar > max_pixels:
        beta = math.sqrt((height * width) / max_pixels)
        h_bar = max(factor, math.floor(height / beta / factor) * factor)
        w_bar = max(factor, math.floor(width / beta / factor) * factor)
    elif h_bar * w_bar < min_pixels:
        beta = math.sqrt(min_pixels / (height * width))
        h_bar = math.ceil(height * beta / factor) * factor
        w_bar = math.ceil(width * beta / factor) * factor

    return h_bar, w_bar


def patchify(array: np.ndarray, patch_size: int, batched: bool = False) -> np.ndarray:
    """Reshape image(s) to patches."""
    if len(array.shape) == 3 and not batched:
        h, w, c = array.shape
        h_patches = h // patch_size
        w_patches = w // patch_size
        array = array.reshape(h_patches, patch_size, w_patches, patch_size, c)
        array = array.transpose(0, 2, 1, 3, 4)
        return array.reshape(h_patches * w_patches, patch_size * patch_size * c)
    elif len(array.shape) == 4 or (len(array.shape) == 3 and batched):
        if len(array.shape) == 3:
            bs, h, w = array.shape
            c = 1
            array = array[..., None]
        else:
            bs, h, w, c = array.shape
        h_patches = h // patch_size
        w_patches = w // patch_size
        array = array.reshape(bs, h_patches, patch_size, w_patches, patch_size, c)
        array = array.transpose(0, 1, 3, 2, 4, 5)
        result = array.reshape(bs, h_patches * w_patches, patch_size * patch_size * c)
        if c == 1:
            result = result[..., 0] if result.shape[-1] == 1 else result.mean(axis=-1)
        return result
    else:
        raise ValueError(f"Unsupported array shape: {array.shape}")


class ImageProcessor:
    """Image processor for Jina VLM (standalone, not a BaseImageProcessor)."""

    def __init__(
        self,
        config: Optional[dict] = None,
        base_input_size: Tuple[int, int] = (378, 378),
        patch_size: int = 14,
        max_crops: int = 12,
        min_pixels: int = 3136,
        max_pixels: int = 1003520,
        overlap_margins: Tuple[int, int] = (4, 4),
        pooling_h: int = 2,
        pooling_w: int = 2,
        use_column_tokens: bool = True,
        image_mean: Optional[List[float]] = None,
        image_std: Optional[List[float]] = None,
        patch_token_id: int = DEFAULT_PATCH_TOKEN_ID,
        start_token_id: int = DEFAULT_START_TOKEN_ID,
        end_token_id: int = DEFAULT_END_TOKEN_ID,
        column_token_id: int = DEFAULT_COLUMN_TOKEN_ID,
    ):
        self.base_input_size = base_input_size
        self.patch_size = patch_size
        self.max_crops = max_crops
        self.min_pixels = min_pixels
        self.max_pixels = max_pixels
        self.overlap_margins = overlap_margins
        self.pooling_h = pooling_h
        self.pooling_w = pooling_w
        self.use_column_tokens = use_column_tokens
        self.image_mean = image_mean or CLIP_MEAN
        self.image_std = image_std or CLIP_STD

        self.patch_token_id = patch_token_id
        self.start_token_id = start_token_id
        self.end_token_id = end_token_id
        self.column_token_id = column_token_id

        self.crop_patches = base_input_size[0] // patch_size
        self.token_length_h = (self.crop_patches + pooling_h - 1) // pooling_h
        self.token_length_w = (self.crop_patches + pooling_w - 1) // pooling_w
        self.tokens_per_image = self.token_length_h * self.token_length_w

    def normalize(self, x: np.ndarray) -> np.ndarray:
        return (x - 0.5) * 2.0

    def resize_image(
        self,
        image: np.ndarray,
        size: Tuple[int, int],
    ) -> Tuple[np.ndarray, np.ndarray]:
        pil_image = Image.fromarray((image * 255).astype(np.uint8))
        pil_image = pil_image.resize((size[1], size[0]), Image.BICUBIC)
        resized = np.array(pil_image, dtype=np.float32) / 255.0
        mask = np.ones((size[0], size[1]), dtype=np.bool_)
        return resized, mask

    def select_tiling(
        self,
        h: int,
        w: int,
        patch_size: int,
        max_crops: int,
    ) -> Tuple[int, int]:
        tilings = []
        for i in range(1, max_crops + 1):
            for j in range(1, max_crops + 1):
                if i * j <= max_crops:
                    tilings.append((i, j))

        tilings.sort(key=lambda x: (x[0] * x[1], x[0]))
        candidate_tilings = np.array(tilings, dtype=np.int32)
        candidate_resolutions = candidate_tilings * patch_size

        original_size = np.array([h, w], dtype=np.float32)
        with np.errstate(divide="ignore"):
            required_scale = candidate_resolutions.astype(np.float32) / original_size
        required_scale = np.min(required_scale, axis=-1, keepdims=True)

        if np.all(required_scale < 1):
            ix = np.argmax(required_scale)
        else:
            required_scale = np.where(required_scale < 1.0, 1e9, required_scale)
            ix = np.argmin(required_scale)

        return tuple(candidate_tilings[ix])

    def _get_patches_from_tiling(
        self,
        num_tiles: int,
        pooling_size: int,
        crop_patches: int,
        crop_window_patches: int,
        left_margin: int,
        right_margin: int,
    ) -> int:
        if num_tiles > 1:
            left_crop = (
                (crop_window_patches + left_margin + pooling_size - 1)
                // pooling_size
                * pooling_size
            )
            middle_crop = (
                (crop_window_patches + pooling_size - 1) // pooling_size * pooling_size
            )
            right_crop = (
                (crop_window_patches + right_margin + pooling_size - 1)
                // pooling_size
                * pooling_size
            )
            return left_crop + (num_tiles - 2) * middle_crop + right_crop
        else:
            return (crop_patches + pooling_size - 1) // pooling_size * pooling_size

    def build_image_input_idx(
        self,
        image_tokens: np.ndarray,
        patch_order: Optional[np.ndarray],
    ) -> np.ndarray:
        image_input_idx = image_tokens == self.patch_token_id
        image_input_idx = np.nonzero(image_input_idx)[0].astype(np.int32)

        if patch_order is not None:
            patch_order = np.reshape(patch_order, [-1])
            valid = patch_order >= 0
            n_valid_patches = valid.sum()

            if len(image_input_idx) != n_valid_patches:
                raise ValueError(
                    f"Mismatch: {len(image_input_idx)} patch tokens but {n_valid_patches} valid patches"
                )

            sorted_patch_ixs = np.zeros([image_input_idx.shape[0]], np.int32)
            sorted_patch_ixs[patch_order[valid]] = np.arange(
                n_valid_patches, dtype=np.int32
            )
            sorted_patch_ixs_ex = np.full(np.shape(patch_order), -1)
            sorted_patch_ixs_ex[valid] = sorted_patch_ixs

            valid_mask = (sorted_patch_ixs_ex >= 0).astype(np.int32)
            image_input_idx = image_input_idx[sorted_patch_ixs_ex * valid_mask]
            image_input_idx = image_input_idx * valid_mask - 10000 * (1 - valid_mask)

        return np.reshape(image_input_idx, [-1, self.tokens_per_image])

    def crop_image(
        self, image: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        left_margin, right_margin = self.overlap_margins
        total_margin_pixels = self.patch_size * (right_margin + left_margin)
        crop_patches = self.crop_patches
        crop_window_patches = crop_patches - (right_margin + left_margin)
        crop_window_size = crop_window_patches * self.patch_size

        original_h, original_w = image.shape[:2]

        tiling = self.select_tiling(
            original_h - total_margin_pixels,
            original_w - total_margin_pixels,
            crop_window_size,
            self.max_crops,
        )

        target_h = tiling[0] * crop_window_size + total_margin_pixels
        target_w = tiling[1] * crop_window_size + total_margin_pixels
        src, img_mask = self.resize_image(image, (target_h, target_w))
        src = self.normalize(src)

        patches_arr = []
        mask_arr = []
        patch_ordering_arr = []

        crop_size = self.base_input_size[0]
        on = 0

        for i in range(tiling[0]):
            y0 = i * crop_window_size
            crop_y0 = 0 if i == 0 else left_margin // self.pooling_h
            crop_h = crop_patches - (right_margin + left_margin)
            if i == 0:
                crop_h += left_margin
            if i == (tiling[0] - 1):
                crop_h += right_margin

            for j in range(tiling[1]):
                x0 = j * crop_window_size
                crop_x0 = 0 if j == 0 else left_margin // self.pooling_w
                crop_w = crop_patches - (right_margin + left_margin)
                if j == 0:
                    crop_w += left_margin
                if j == (tiling[1] - 1):
                    crop_w += right_margin

                pooled_w = (crop_w + self.pooling_w - 1) // self.pooling_w
                pooled_h = (crop_h + self.pooling_h - 1) // self.pooling_h
                after_padding_width = self.token_length_w - pooled_w - crop_x0
                after_padding_height = self.token_length_h - pooled_h - crop_y0

                patch_ordering_arr.append(
                    np.pad(
                        np.reshape(
                            np.arange(on, on + pooled_h * pooled_w, dtype=np.int32),
                            (pooled_h, pooled_w),
                        ),
                        [
                            [crop_y0, after_padding_height],
                            [crop_x0, after_padding_width],
                        ],
                        constant_values=-1,
                        mode="constant",
                    )
                )

                crop = src[y0 : y0 + crop_size, x0 : x0 + crop_size]
                if crop.shape[0] < crop_size or crop.shape[1] < crop_size:
                    padded = np.zeros((crop_size, crop_size, 3), dtype=np.float32)
                    padded[: crop.shape[0], : crop.shape[1]] = crop
                    crop = padded
                patches_arr.append(crop)

                crop_mask = img_mask[y0 : y0 + crop_size, x0 : x0 + crop_size]
                if crop_mask.shape[0] < crop_size or crop_mask.shape[1] < crop_size:
                    padded_mask = np.zeros((crop_size, crop_size), dtype=np.bool_)
                    padded_mask[: crop_mask.shape[0], : crop_mask.shape[1]] = crop_mask
                    crop_mask = padded_mask
                mask_arr.append(crop_mask)

                on += pooled_h * pooled_w

        patches = np.stack(patches_arr)
        patch_ordering = np.stack(patch_ordering_arr)
        img_masks = np.stack(mask_arr)

        patches = patchify(patches, self.patch_size, batched=True)
        img_masks = patchify(
            img_masks.astype(np.float32), self.patch_size, batched=True
        )
        if img_masks.ndim == 3:
            img_masks = img_masks.mean(axis=-1)

        patch_ordering = np.reshape(patch_ordering, [-1])
        valid = patch_ordering >= 0

        patch_ordering_rh = np.reshape(
            patch_ordering,
            [tiling[0], tiling[1], self.token_length_h, self.token_length_w],
        )
        patch_ordering_rh = np.transpose(patch_ordering_rh, [0, 2, 1, 3])
        patch_ordering_rh = np.reshape(patch_ordering_rh, [-1])
        patch_ordering[valid] = patch_ordering_rh[patch_ordering_rh >= 0]

        h = self._get_patches_from_tiling(
            tiling[0],
            self.pooling_h,
            crop_patches,
            crop_window_patches,
            left_margin,
            right_margin,
        )
        w = self._get_patches_from_tiling(
            tiling[1],
            self.pooling_w,
            crop_patches,
            crop_window_patches,
            left_margin,
            right_margin,
        )

        per_row = np.full((w // self.pooling_w,), self.patch_token_id, dtype=np.int32)
        if self.use_column_tokens:
            per_row = np.concatenate([per_row, [self.column_token_id]], 0)
        joint = np.tile(per_row, [h // self.pooling_h])
        joint = [[self.start_token_id], joint, [self.end_token_id]]

        thumb, _ = self.resize_image(image, self.base_input_size)
        thumb = self.normalize(thumb)
        thumb_patches = patchify(thumb, self.patch_size, batched=False)
        patches = np.concatenate([np.expand_dims(thumb_patches, 0), patches], 0)

        patch_ordering = np.where(
            patch_ordering >= 0, patch_ordering + self.tokens_per_image, -1
        )
        patch_ordering = np.concatenate(
            [np.arange(0, self.tokens_per_image), patch_ordering], 0
        )

        per_row = np.full((self.token_length_w,), self.patch_token_id, dtype=np.int32)
        if self.use_column_tokens:
            per_row = np.concatenate([per_row, [self.column_token_id]], 0)
        extra_tokens = np.tile(per_row, [self.token_length_h])
        joint = [[self.start_token_id], extra_tokens, [self.end_token_id]] + joint

        image_tokens = np.concatenate(joint, 0).astype(np.int32)

        img_masks = np.pad(img_masks, [[1, 0], [0, 0]], constant_values=1.0)

        return patches, image_tokens, patch_ordering, img_masks

    def process_image(
        self,
        image: Union[Image.Image, np.ndarray],
    ) -> Dict[str, np.ndarray]:
        if isinstance(image, Image.Image):
            image = image.convert("RGB")
            image = np.array(image, dtype=np.float32) / 255.0
        elif image.dtype == np.uint8:
            image = image.astype(np.float32) / 255.0

        h, w = image.shape[:2]
        new_h, new_w = smart_resize(
            h,
            w,
            factor=self.patch_size,
            min_pixels=self.min_pixels,
            max_pixels=self.max_pixels,
        )
        if (new_h, new_w) != (h, w):
            image, _ = self.resize_image(image, (new_h, new_w))

        patches, image_tokens, patch_ordering, masks = self.crop_image(image)

        image_input_idx = self.build_image_input_idx(image_tokens, patch_ordering)

        return {
            "pixel_values": patches,
            "image_tokens": image_tokens,
            "image_input_idx": image_input_idx,
            "image_masks": masks,
        }

    def preprocess(
        self,
        images: Union[Image.Image, List[Image.Image], str, List[str]],
        **kwargs,
    ) -> Dict[str, mx.array]:
        if not isinstance(images, list):
            images = [images]

        loaded_images = []
        for img in images:
            if isinstance(img, str):
                loaded_images.append(Image.open(img).convert("RGB"))
            else:
                loaded_images.append(img)

        results = {
            "pixel_values": [],
            "image_tokens": [],
            "image_input_idx": [],
            "image_masks": [],
        }

        for image in loaded_images:
            processed = self.process_image(image)
            for key in results:
                results[key].append(processed[key])

        return results



import inspect
from dataclasses import dataclass, field
from typing import Dict, List, Optional, Union

from ..base import BaseModelConfig


@dataclass
class Code2WavConfig(BaseModelConfig):
    model_type: str = ""
    hidden_size: int = 1024
    intermediate_size: int = 3072
    num_hidden_layers: int = 8
    num_attention_heads: int = 16
    num_key_value_heads: int = 16
    decoder_dim: int = 1536
    hidden_act: str = "silu"
    attention_bias: bool = False
    attention_dropout: float = 0.0
    rms_norm_eps: float = 1e-05
    rope_theta: float = 10000
    max_position_embeddings: int = 8000
    sliding_window: Optional[int] = 72
    codebook_dim: int = 512
    codebook_size: int = 2048
    num_quantizers: int = 16
    num_semantic_quantizers: int = 1
    semantic_codebook_size: int = 4096
    vector_quantization_hidden_dimension: int = 512
    layer_scale_initial_scale: float = 0.01
    upsample_rates: List[int] = field(default_factory=lambda: [8, 5, 4, 3])
    upsampling_ratios: List[int] = field(default_factory=lambda: [2, 2])


@dataclass
class AudioConfig(BaseModelConfig):
    model_type: str = "qwen3_omni_moe_audio_encoder"
    d_model: int = 1280
    encoder_layers: int = 32
    encoder_attention_heads: int = 20
    encoder_ffn_dim: int = 5120
    num_hidden_layers: int = 32
    num_mel_bins: int = 128
    output_dim: int = 2048
    downsample_hidden_size: int = 480
    activation_function: str = "gelu"
    activation_dropout: float = 0.0
    attention_dropout: float = 0.0
    dropout: float = 0.0
    initializer_range: float = 0.02
    scale_embedding: bool = False
    conv_chunksize: int = 500
    n_window: int = 50
    n_window_infer: int = 800
    max_source_positions: int = 1500


@dataclass
class VisionConfig(BaseModelConfig):
    model_type: str = "qwen3_omni_moe_vision_encoder"
    depth: int = 27
    hidden_size: int = 1152
    intermediate_size: int = 4304
    out_hidden_size: int = 2048
    num_heads: int = 16
    image_size: int = 768
    patch_size: int = 16
    spatial_patch_size: int = 16
    spatial_merge_size: int = 2
    temporal_patch_size: int = 2
    tokens_per_second: int = 2
    in_channels: int = 3
    in_chans: int = 3
    hidden_act: str = "gelu_pytorch_tanh"
    initializer_range: float = 0.02
    apply_vit_abs_pos_embed: bool = True
    num_position_embeddings: int = 2304
    deepstack_visual_indexes: List[int] = field(default_factory=lambda: [8, 16, 24])


@dataclass
class TextConfig(BaseModelConfig):
    num_hidden_layers: int
    hidden_size: int
    intermediate_size: int
    num_attention_heads: int
    num_experts: int
    num_experts_per_tok: int
    decoder_sparse_step: int
    mlp_only_layers: List[int]
    moe_intermediate_size: int
    rms_norm_eps: float
    vocab_size: int
    num_key_value_heads: Optional[int]
    head_dim: int
    rope_theta: float
    max_position_embeddings: int
    model_type: str = "qwen3_omni_moe_text_encoder"
    norm_topk_prob: bool = True
    rope_scaling: Optional[Dict[str, Union[float, str, bool, List[int]]]] = field(
        default_factory=lambda: {"type": "default"}
    )
    tie_word_embeddings: bool = False
    attention_bias: bool = False
    attention_dropout: float = 0.0
    hidden_act: str = "silu"
    use_cache: bool = True
    use_sliding_window: bool = False
    sliding_window: Optional[int] = None
    shared_expert_intermediate_size: Optional[int] = None
    router_aux_loss_coef: float = 0.001
    use_qk_norm: bool = False
    initializer_range: float = 0.02

    def __post_init__(self):
        if self.num_key_value_heads is None:
            self.num_key_value_heads = self.num_attention_heads


@dataclass
class CodePredictorConfig(BaseModelConfig):
    model_type: str = "qwen3_omni_moe_talker_code_predictor"
    num_hidden_layers: int = 5
    hidden_size: int = 1024
    intermediate_size: int = 3072
    num_attention_heads: int = 16
    num_key_value_heads: int = 8
    head_dim: int = 128
    hidden_act: str = "silu"
    rms_norm_eps: float = 1e-06
    rope_theta: float = 1000000
    max_position_embeddings: int = 32768
    vocab_size: int = 2048
    num_code_groups: int = 16
    max_window_layers: int = 28
    attention_bias: bool = False
    attention_dropout: float = 0.0
    initializer_range: float = 0.02
    use_cache: bool = True
    layer_types: List[str] = field(default_factory=lambda: ["full_attention"] * 5)
    rope_scaling: Optional[Dict[str, Union[float, str, bool, List[int]]]] = None
    use_sliding_window: bool = False
    sliding_window: Optional[int] = None


@dataclass
class ThinkerConfig(BaseModelConfig):
    text_config: TextConfig
    vision_config: VisionConfig
    audio_config: AudioConfig
    model_type: str = "qwen3_omni_moe_thinker"
    dtype: str = "bfloat16"
    initializer_range: float = 0.02
    image_token_id: int = 151655
    video_token_id: int = 151656
    audio_token_id: int = 151675
    audio_start_token_id: int = 151669
    audio_end_token_id: int = 151670
    vision_start_token_id: int = 151652
    vision_end_token_id: int = 151653
    user_token_id: int = 872
    position_id_per_seconds: int = 13
    seconds_per_chunk: int = 2

    @classmethod
    def from_dict(cls, params):
        params = dict(params)
        text_config = TextConfig.from_dict(params.pop("text_config", {}))
        vision_config = VisionConfig.from_dict(params.pop("vision_config", {}))
        audio_config = AudioConfig.from_dict(params.pop("audio_config", {}))

        return cls(
            text_config=text_config,
            vision_config=vision_config,
            audio_config=audio_config,
            **{
                k: v
                for k, v in params.items()
                if k in inspect.signature(cls).parameters
            },
        )


@dataclass
class TalkerConfig(BaseModelConfig):
    text_config: TextConfig
    code_predictor_config: CodePredictorConfig
    model_type: str = "qwen3_omni_moe_talker"
    accept_hidden_layer: int = 24
    image_token_id: int = 151655
    video_token_id: int = 151656
    audio_token_id: int = 151675
    audio_start_token_id: int = 151669
    audio_end_token_id: int = 151670
    vision_start_token_id: int = 151652
    num_code_groups: int = 16
    output_router_logits: bool = False
    position_id_per_seconds: int = 13
    seconds_per_chunk: int = 2
    spatial_merge_size: int = 2
    thinker_hidden_size: int = 2048
    codec_bos_id: int = 2149
    codec_eos_token_id: int = 2150
    codec_nothink_id: int = 2155
    codec_pad_id: int = 2148
    codec_think_bos_id: int = 2156
    codec_think_eos_id: int = 2157
    speaker_id: Dict[str, int] = field(
        default_factory=lambda: {"chelsie": 2301, "ethan": 2302, "aiden": 2303}
    )

    @classmethod
    def from_dict(cls, params):
        params = dict(params)
        text_config = TextConfig.from_dict(params.pop("text_config", {}))
        code_predictor_config = CodePredictorConfig.from_dict(
            params.pop("code_predictor_config", {})
        )

        return cls(
            text_config=text_config,
            code_predictor_config=code_predictor_config,
            **{
                k: v
                for k, v in params.items()
                if k in inspect.signature(cls).parameters
            },
        )


@dataclass
class ModelConfig(BaseModelConfig):
    thinker_config: ThinkerConfig
    talker_config: TalkerConfig
    code2wav_config: Code2WavConfig
    model_type: str = "qwen3_omni_moe"
    dtype: str = "bfloat16"
    enable_audio_output: bool = True
    im_start_token_id: int = 151644
    im_end_token_id: int = 151645
    system_token_id: int = 8948
    user_token_id: int = 872
    assistant_token_id: int = 77091
    tts_bos_token_id: int = 151672
    tts_eos_token_id: int = 151673
    tts_pad_token_id: int = 151671

    @classmethod
    def from_dict(cls, params):
        params = dict(params)
        thinker_config = ThinkerConfig.from_dict(params.pop("thinker_config", {}))
        talker_config = TalkerConfig.from_dict(params.pop("talker_config", {}))
        code2wav_config = Code2WavConfig.from_dict(params.pop("code2wav_config", {}))

        return cls(
            thinker_config=thinker_config,
            talker_config=talker_config,
            code2wav_config=code2wav_config,
            **{
                k: v
                for k, v in params.items()
                if k in inspect.signature(cls).parameters
            },
        )



from typing import Dict, Optional

import mlx.core as mx
import mlx.nn as nn

from ..base import InputEmbeddingsFeatures, check_array_shape
from .config import ModelConfig
from .language import LanguageModel
from .vision import VisionModel

try:
    from transformers import AutoImageProcessor, AutoProcessor

    from .processing_hunyuan_vl import HunYuanVLImageProcessor, HunYuanVLProcessor

    MODEL_TYPE = "hunyuan_vl"

    AutoImageProcessor.register(
        MODEL_TYPE, slow_image_processor_class=HunYuanVLImageProcessor
    )
    AutoProcessor.register(MODEL_TYPE, HunYuanVLProcessor)

except Exception as e:
    raise e


class Model(nn.Module):

    def __init__(self, config: ModelConfig):
        super().__init__()
        self.config = config
        self.model_type = config.model_type
        self.vision_tower = VisionModel(config.vision_config)
        self.language_model = LanguageModel(config)

    def get_input_embeddings(
        self,
        input_ids: Optional[mx.array] = None,
        pixel_values: Optional[mx.array] = None,
        **kwargs,
    ) -> mx.array:

        image_grid_thw = kwargs.pop("image_grid_thw", None)

        position_ids_from_processor = kwargs.pop("position_ids", None)

        # Get text embeddings
        inputs_embeds = self.language_model.model.embed_tokens(input_ids)

        # If no image, return text embeddings
        if pixel_values is None:
            # Reset stored position_ids when no image
            self.language_model._position_ids = None
            return InputEmbeddingsFeatures(inputs_embeds=inputs_embeds)

        # Get vision features
        vision_features = self.vision_tower(pixel_values, image_grid_thw)

        # Find image token positions and replace with vision features
        image_token_id = self.config.image_token_id
        image_mask = input_ids == image_token_id

        # Get number of image tokens expected
        num_image_tokens = image_mask.sum().item()
        num_vision_tokens = vision_features.shape[1]

        if num_image_tokens != num_vision_tokens:
            raise ValueError(
                f"Number of image placeholders ({num_image_tokens}) does not match "
                f"number of vision tokens ({num_vision_tokens}). "
                f"Expected token count based on grid: {num_vision_tokens}"
            )

        B, L, _ = inputs_embeds.shape

        output_parts = []

        for b in range(B):
            mask_b = image_mask[b]  # (L,) boolean mask
            text_embeds_b = inputs_embeds[b]  # (L, D)
            vis_feats_b = vision_features[b]  # (num_vis_tokens, D)

            # Build sequence for this batch
            vis_idx = 0
            seq_parts = []
            for pos in range(L):
                if mask_b[pos].item():
                    # Use vision feature
                    seq_parts.append(vis_feats_b[vis_idx : vis_idx + 1])
                    vis_idx += 1
                else:
                    # Use text embedding
                    seq_parts.append(text_embeds_b[pos : pos + 1])

            # Concatenate all parts for this batch
            batch_embeds = mx.concatenate(seq_parts, axis=0)  # (L, D)
            output_parts.append(batch_embeds[None, :, :])  # (1, L, D)

        # Stack batches
        inputs_embeds = mx.concatenate(output_parts, axis=0)  # (B, L, D)

        # Pre-calculate position_ids for chunked prefill
        if position_ids_from_processor is not None:
            self.language_model._position_ids = position_ids_from_processor
        elif image_grid_thw is not None:
            position_ids = self.language_model.get_xdrope_input_positions(
                input_tokens=input_ids[0].tolist(),
                image_grid_thw=image_grid_thw,
                image_token_id=self.config.image_token_id,
                spatial_merge_size=self.config.vision_config.spatial_merge_size,
            )[None, ...]
            self.language_model._position_ids = position_ids

        return InputEmbeddingsFeatures(inputs_embeds=inputs_embeds)

    @property
    def layers(self):
        return self.language_model.model.layers

    @property
    def head_dim(self):
        return self.config.text_config.head_dim

    @property
    def n_kv_heads(self):
        return self.config.text_config.num_key_value_heads

    def __call__(
        self,
        input_ids: mx.array,
        pixel_values: Optional[mx.array] = None,
        mask: Optional[mx.array] = None,
        cache=None,
        **kwargs,
    ):

        # Get embeddings (with vision features merged if image provided)
        input_embeddings_features = self.get_input_embeddings(
            input_ids=input_ids,
            pixel_values=pixel_values,
            **kwargs,
        )

        # Forward through language model
        return self.language_model(
            input_ids=input_ids,
            inputs_embeds=input_embeddings_features.inputs_embeds,
            mask=mask,
            cache=cache,
            image_grid_thw=image_grid_thw,
        )

    def sanitize(self, weights: Dict[str, mx.array]) -> Dict[str, mx.array]:

        sanitized = {}

        for key, value in weights.items():
            new_key = key

            # Language model mappings
            if key.startswith("model."):
                new_key = "language_model." + key

            # Vision tower mappings
            elif key.startswith("vit."):
                new_key = key.replace("vit.", "vision_tower.", 1)

            # Handle Conv2d weight transposition for MLX
            # PyTorch Conv2d: [out_channels, in_channels, kH, kW]
            # MLX Conv2d: [out_channels, kH, kW, in_channels]
            if (
                "patch_embedding.weight" in new_key
                or "proj.0.weight" in new_key
                or "proj.2.weight" in new_key
            ):
                if not check_array_shape(value):
                    value = value.transpose(0, 2, 3, 1)

            sanitized[new_key] = value

        return sanitized



from typing import Optional, Tuple

import mlx.core as mx
import mlx.nn as nn

from ..base import InputEmbeddingsFeatures
from ..cache import KVCache
from .config import ModelConfig
from .language import LanguageModel
from .vision import VisionModel


class Model(nn.Module):
    def __init__(self, config: ModelConfig):
        super().__init__()
        self.config = config
        self.vision_tower = VisionModel(config.vision_config)
        self.language_model = LanguageModel(config.text_config)
        self.multi_modal_projector = nn.Linear(
            config.vision_config.vision_output_dim,
            config.text_config.hidden_size,
            bias=True,
        )

    @property
    def layers(self):
        return self.language_model.model.layers

    def get_input_embeddings(
        self,
        input_ids: Optional[mx.array] = None,
        pixel_values: Optional[mx.array] = None,
        **kwargs,
    ):
        aspect_ratio_ids = kwargs.get("aspect_ratio_ids", None)
        aspect_ratio_mask = kwargs.get("aspect_ratio_mask", None)
        cross_attention_mask = kwargs.get("cross_attention_mask", None)

        # Get text embeddings
        inputs_embeds = self.language_model.model.embed_tokens(input_ids)

        cross_attention_states = None
        full_text_row_masked_out_mask = None

        # Process vision input if provided
        if pixel_values is not None:
            if aspect_ratio_ids is None:
                raise ValueError(
                    "`aspect_ratio_ids` must be provided if `pixel_values` is provided"
                )

            vision_outputs = self.vision_tower(
                pixel_values=pixel_values,
                aspect_ratio_ids=aspect_ratio_ids,
                aspect_ratio_mask=aspect_ratio_mask,
            )
            cross_attention_states = vision_outputs[0]

            cross_attention_states = self.multi_modal_projector(
                cross_attention_states
            ).reshape(
                -1,
                cross_attention_states.shape[-2],
                self.config.text_config.hidden_size,
            )

        # Prepare cross attention mask
        if cross_attention_mask is not None:
            cross_attention_mask, full_text_row_masked_out_mask = (
                self._prepare_cross_attention_mask(
                    cross_attention_mask,
                    num_vision_tokens=(
                        self.config.vision_config.image_size
                        // self.config.vision_config.patch_size
                    )
                    ** 2
                    + 1,
                )
            )

            cache_position = mx.arange(input_ids.shape[1], dtype=mx.int32)
            cross_attention_mask = cross_attention_mask[:, :, cache_position]
            full_text_row_masked_out_mask = full_text_row_masked_out_mask[
                :, :, cache_position
            ]

        return InputEmbeddingsFeatures(
            inputs_embeds=inputs_embeds,
            cross_attention_states=cross_attention_states,
            cross_attention_mask=cross_attention_mask,
            full_text_row_masked_out_mask=full_text_row_masked_out_mask,
        )

    def __call__(
        self,
        input_ids: mx.array,
        pixel_values: Optional[mx.array] = None,
        mask: Optional[mx.array] = None,
        cache: Optional[KVCache] = None,
        **kwargs,
    ) -> Tuple[mx.array, Optional[mx.array]]:

        aspect_ratio_ids = kwargs.pop("aspect_ratio_ids", None)
        aspect_ratio_mask = kwargs.pop("aspect_ratio_mask", None)
        cross_attention_mask = kwargs.pop("cross_attention_mask", None)

        inputs_embeds = None

        # Process vision input if provided
        if pixel_values is not None:
            if aspect_ratio_ids is None:
                raise ValueError(
                    "`aspect_ratio_ids` must be provided if `pixel_values` is provided"
                )

            vision_outputs = self.vision_tower(
                pixel_values=pixel_values,
                aspect_ratio_ids=aspect_ratio_ids,
                aspect_ratio_mask=aspect_ratio_mask,
            )
            cross_attention_states = vision_outputs[0]

            cross_attention_states = self.multi_modal_projector(
                cross_attention_states
            ).reshape(
                -1,
                cross_attention_states.shape[-2],
                self.config.text_config.hidden_size,
            )

        else:
            cross_attention_states = None

        # Prepare cross attention mask
        if cross_attention_mask is not None:
            cross_attention_mask, full_text_row_masked_out_mask = (
                self._prepare_cross_attention_mask(
                    cross_attention_mask,
                    num_vision_tokens=(
                        self.config.vision_config.image_size
                        // self.config.vision_config.patch_size
                    )
                    ** 2
                    + 1,
                )
            )
        else:
            full_text_row_masked_out_mask = None

        if cross_attention_mask is not None:
            cache_position = mx.arange(input_ids.shape[1], dtype=mx.int32)
            cross_attention_mask = cross_attention_mask[:, :, cache_position]
            full_text_row_masked_out_mask = full_text_row_masked_out_mask[
                :, :, cache_position
            ]

        # Process language input
        outputs = self.language_model(
            inputs=input_ids,
            mask=mask,
            cross_attention_states=cross_attention_states,
            cross_attention_mask=cross_attention_mask,
            full_text_row_masked_out_mask=full_text_row_masked_out_mask,
            inputs_embeds=inputs_embeds,
            cache=cache,
        )

        return outputs

    def _prepare_cross_attention_mask(
        self,
        cross_attention_mask: mx.array,
        num_vision_tokens: int,
    ) -> Tuple[mx.array, mx.array]:
        batch_size, text_total_length, *_ = cross_attention_mask.shape
        cross_attention_mask = mx.repeat(
            cross_attention_mask, num_vision_tokens, axis=3
        )
        cross_attention_mask = cross_attention_mask.reshape(
            batch_size, text_total_length, -1
        )
        cross_attention_mask = mx.expand_dims(cross_attention_mask, 1)

        # Invert the mask
        inverted_cross_attn_mask = 1.0 - cross_attention_mask
        fill_array = mx.array(-1e9)
        fill_array = mx.broadcast_to(fill_array, inverted_cross_attn_mask.shape)
        cross_attention_mask = mx.where(
            inverted_cross_attn_mask,
            fill_array,
            cross_attention_mask,
        )

        # Apply full-row bias
        full_text_row_masked_out_mask = mx.any(
            cross_attention_mask != -1e9,
            axis=-1,
            keepdims=True,
        )
        cross_attention_mask *= full_text_row_masked_out_mask

        return cross_attention_mask, full_text_row_masked_out_mask

    def sanitize(self, weights):
        def transform_key(key):
            if "vision_tower" not in key:
                key = key.replace("vision_model", "vision_tower")
            return key

        return {transform_key(k): v for k, v in weights.items()}



import argparse
import codecs
import contextlib
import functools
import json
import time
from dataclasses import dataclass
from typing import Any, Callable, Dict, Generator, List, Optional, Tuple, Union

import mlx.core as mx
import mlx.nn as nn
from mlx.utils import tree_reduce
from mlx_lm.generate import maybe_quantize_kv_cache
from mlx_lm.sample_utils import make_logits_processors, make_sampler
from tqdm import tqdm
from transformers import PreTrainedTokenizer

from .models import cache
from .prompt_utils import apply_chat_template
from .utils import StoppingCriteria, group_images_by_shape, load, prepare_inputs

DEFAULT_MODEL_PATH = "mlx-community/nanoLLaVA-1.5-8bit"
DEFAULT_IMAGE = None
DEFAULT_AUDIO = None
DEFAULT_PROMPT = "What are these?"
DEFAULT_MAX_TOKENS = 256
DEFAULT_TEMPERATURE = 0.5
DEFAULT_TOP_P = 1.0
DEFAULT_SEED = 0
DEFAULT_QUANTIZED_KV_START = 5000


def parse_arguments():
    parser = argparse.ArgumentParser(
        description="Generate text from an image using a model."
    )
    parser.add_argument(
        "--model",
        type=str,
        default=DEFAULT_MODEL_PATH,
        help="The path to the local model directory or Hugging Face repo.",
    )
    parser.add_argument(
        "--adapter-path",
        type=str,
        default=None,
        help="The path to the adapter weights.",
    )
    parser.add_argument(
        "--image",
        type=str,
        nargs="+",
        default=DEFAULT_IMAGE,
        help="URL or path of the image to process.",
    )
    parser.add_argument(
        "--audio",
        type=str,
        nargs="+",
        default=DEFAULT_AUDIO,
        help="URL or path of the audio to process.",
    )
    parser.add_argument(
        "--resize-shape",
        type=int,
        nargs="+",
        default=None,
        help="Resize shape for the image.",
    )
    parser.add_argument(
        "--prompt",
        type=str,
        nargs="+",
        default=DEFAULT_PROMPT,
        help="Message to be processed by the model.",
    )
    parser.add_argument(
        "--system",
        type=str,
        default=None,
        help="System message for the model.",
    )
    parser.add_argument(
        "--max-tokens",
        type=int,
        default=DEFAULT_MAX_TOKENS,
        help="Maximum number of tokens to generate.",
    )
    parser.add_argument(
        "--temperature",
        type=float,
        default=DEFAULT_TEMPERATURE,
        help="Temperature for sampling.",
    )
    parser.add_argument("--chat", action="store_true", help="Chat in multi-turn style.")
    parser.add_argument("--verbose", action="store_false", help="Detailed output.")
    parser.add_argument(
        "--eos-tokens",
        type=str,
        nargs="+",
        default=None,
        help="EOS tokens to add to the tokenizer.",
    )
    parser.add_argument(
        "--max-kv-size",
        type=int,
        default=None,
        help="Maximum KV size for the prompt cache.",
    )
    parser.add_argument(
        "--kv-bits",
        type=int,
        default=None,
        help="Number of bits to quantize the KV cache to.",
    )
    parser.add_argument(
        "--kv-group-size",
        type=int,
        default=64,
        help="Group size for the KV cache.",
    )
    parser.add_argument(
        "--quantized-kv-start",
        type=int,
        default=DEFAULT_QUANTIZED_KV_START,
        help="Start index for the quantized KV cache.",
    )
    parser.add_argument(
        "--skip-special-tokens",
        action="store_true",
        help="Skip special tokens in the detokenizer.",
    )
    parser.add_argument(
        "--force-download",
        action="store_true",
        help="Force download the model from Hugging Face.",
    )
    parser.add_argument(
        "--revision",
        type=str,
        default="main",
        help="The specific model version to use (branch, tag, commit).",
    )
    parser.add_argument(
        "--trust-remote-code",
        action="store_true",
        help="Trust remote code when loading the model.",
    )
    parser.add_argument(
        "--quantize-activations",
        "-qa",
        action="store_true",
        help="Enable activation quantization for QQLinear layers. "
        "Only supported for models quantized with 'nvfp4' or 'mxfp8' modes.",
    )
    parser.add_argument(
        "--processor-kwargs",
        type=json.loads,
        default={},
        help="Extra processor kwargs as JSON. "
        'Example: --processor-kwargs \'{"cropping": false, "max_patches": 3}\'',
    )
    parser.add_argument(
        "--prefill-step-size",
        type=int,
        default=None,
        help="Number of tokens to process per prefill step. "
        "Lower values reduce peak memory usage but may be slower. "
        "Try 512 or 256 if you hit GPU memory errors during prefill.",
    )

    return parser.parse_args()


# A stream on the default device just for generation
generation_stream = mx.new_stream(mx.default_device())


@contextlib.contextmanager
def wired_limit(model: nn.Module, streams: Optional[List[mx.Stream]] = None):
    """
    A context manager to temporarily change the wired limit.

    Note, the wired limit should not be changed during an async eval.  If an
    async eval could be running pass in the streams to synchronize with prior
    to exiting the context manager.
    """
    if not mx.metal.is_available():
        yield
        return

    model_bytes = tree_reduce(
        lambda acc, x: acc + x.nbytes if isinstance(x, mx.array) else acc, model, 0
    )
    max_rec_size = mx.device_info()["max_recommended_working_set_size"]
    if model_bytes > 0.9 * max_rec_size:
        model_mb = model_bytes // 2**20
        max_rec_mb = max_rec_size // 2**20
        print(
            f"[WARNING] Generating with a model that requires {model_mb} MB "
            f"which is close to the maximum recommended size of {max_rec_mb} "
            "MB. This can be slow. See the documentation for possible work-arounds: "
            "https://github.com/ml-explore/mlx-lm/tree/main#large-models"
        )
    old_limit = mx.set_wired_limit(max_rec_size)
    try:
        yield
    finally:
        if streams is not None:
            for s in streams:
                mx.synchronize(s)
        else:
            mx.synchronize()
        mx.set_wired_limit(old_limit)


@dataclass
class GenerationResult:
    text: str = ""
    token: Optional[int] = None
    logprobs: Optional[List[float]] = None
    prompt_tokens: int = 0
    generation_tokens: int = 0
    total_tokens: int = 0
    prompt_tps: float = 0.0
    generation_tps: float = 0.0
    peak_memory: float = 0.0


def generate_step(
    input_ids: mx.array,
    model: nn.Module,
    pixel_values,
    mask,
    *,
    max_tokens: int = 256,
    temperature: float = 0.0,
    repetition_penalty: Optional[float] = None,
    repetition_context_size: Optional[int] = 20,
    top_p: float = 1.0,
    logit_bias: Optional[Dict[int, float]] = None,
    prompt_cache: Optional[List[Any]] = None,
    max_kv_size: Optional[int] = None,
    kv_bits: Optional[int] = None,
    kv_group_size: int = 64,
    quantized_kv_start: int = 0,
    sampler: Optional[Callable[[mx.array], mx.array]] = None,
    logits_processors: Optional[List[Callable[[mx.array, mx.array], mx.array]]] = None,
    prefill_step_size: Optional[int] = 2048,
    **kwargs,
) -> Generator[Tuple[mx.array, mx.array], None, None]:
    """
    A generator producing token ids based on the given prompt from the model.

    Args:
        input_ids (mx.array): The input prompt token ids.
        model (nn.Module): The model to use for generation.
        pixel_values: The pixel values for vision models (optional).
        mask: The attention mask (optional).
        max_tokens (int): Maximum number of tokens to generate. Default: ``256``.
        temperature (float): The temperature for sampling, if 0 the argmax is used.
          Default: ``0``.
        repetition_penalty (float, optional): The penalty factor for repeating
          tokens.
        repetition_context_size (int, optional): The number of tokens to
          consider for repetition penalty. Default: ``20``.
        top_p (float, optional): Nucleus sampling, higher means model considers
          more less likely words.
        logit_bias (dictionary, optional): Additive logit bias.
        prompt_cache (list, optional): Pre-existing KV cache for the prompt.
        max_kv_size (int, optional): Maximum KV cache size.
        kv_bits (int, optional): Number of bits for KV cache quantization.
        kv_group_size (int): Group size for KV cache quantization. Default: ``64``.
        quantized_kv_start (int): Start index for quantized KV cache. Default: ``0``.
        sampler (Callable[mx.array, mx.array], optional): A sampler for sampling a
          token from a vector of log probabilities. Default: ``None``.
        logits_processors (List[Callable[[mx.array, mx.array], mx.array]], optional):
          A list of functions that take tokens and logits and return the processed
          logits. Default: ``None``.
        prefill_step_size (int): Number of tokens to process per prefill step.
          Chunked prefill processes prompts in smaller chunks to reduce peak
          memory usage. Default: ``2048``.

    Yields:
        Generator[Tuple[mx.array, mx.array], None, None]: A generator producing
          one token and a vector of log probabilities.
    """

    quantize_cache_fn = functools.partial(
        maybe_quantize_kv_cache,
        quantized_kv_start=quantized_kv_start,
        kv_group_size=kv_group_size,
        kv_bits=kv_bits,
    )

    if sampler is None:
        sampler = make_sampler(temperature, top_p)

    processors = make_logits_processors(
        logit_bias, repetition_penalty, repetition_context_size
    )
    if logits_processors is not None:
        processors.extend(logits_processors)

    y = input_ids
    tokens = mx.array([], dtype=input_ids.dtype)

    # Create the KV cache for generation
    if prompt_cache is None:
        prompt_cache = cache.make_prompt_cache(
            model.language_model,
            max_kv_size=max_kv_size,
        )

    def _step(y, inputs_embeds=None):
        nonlocal tokens, kwargs

        with mx.stream(generation_stream):
            if "decoder_input_ids" in kwargs:
                outputs = model.language_model(
                    cache=prompt_cache,
                    **kwargs,
                )
            else:
                outputs = model.language_model(
                    y,
                    inputs_embeds=inputs_embeds,
                    cache=prompt_cache,
                    **kwargs,
                )

            logits = outputs.logits[:, -1, :]

            if len(processors) > 0 and len(y) > 0:
                tokens = mx.concat([tokens, y.flatten()])

                for processor in processors:
                    logits = processor(tokens, logits)

            quantize_cache_fn(prompt_cache)

            logprobs = logits - mx.logsumexp(logits)
            y = sampler(logprobs)

            if outputs.cross_attention_states is not None:
                kwargs = {"cross_attention_states": outputs.cross_attention_states}
            elif outputs.encoder_outputs is not None:
                kwargs = {"encoder_outputs": outputs.encoder_outputs}
            else:
                kwargs = {}

            return y, logprobs.squeeze(0)

    with mx.stream(generation_stream):

        # Get input embeddings (handles both multimodal and text-only)
        embedding_output = model.get_input_embeddings(
            input_ids, pixel_values, mask=mask, **kwargs
        )

        inputs_embeds = embedding_output.inputs_embeds

        kwargs.update(
            {
                k: v
                for k, v in embedding_output.to_dict().items()
                if k != "inputs_embeds" and v is not None
            }
        )
        if prefill_step_size is not None and inputs_embeds.shape[1] > prefill_step_size:
            # Chunked prefill with embeddings
            total_tokens = inputs_embeds.shape[1]
            with tqdm(total=total_tokens, desc="Prefill", unit="tok") as pbar:
                while inputs_embeds.shape[1] > 1:
                    n_to_process = min(prefill_step_size, inputs_embeds.shape[1] - 1)
                    model.language_model(
                        inputs=input_ids[:, :n_to_process],
                        inputs_embeds=inputs_embeds[:, :n_to_process],
                        cache=prompt_cache,
                        **kwargs,
                    )
                    quantize_cache_fn(prompt_cache)
                    mx.eval([c.state for c in prompt_cache])
                    inputs_embeds = inputs_embeds[:, n_to_process:]
                    input_ids = input_ids[:, n_to_process:]
                    mx.clear_cache()
                    pbar.update(n_to_process)

            input_ids = input_ids[:, -1:]

        y, logprobs = _step(input_ids, inputs_embeds=inputs_embeds)

    mx.async_eval(y)

    n = 0
    while True:
        if n != max_tokens:
            next_y, next_logprobs = _step(y[None])
            mx.async_eval(next_y)
        if n == 0:
            mx.eval(y)
        if n == max_tokens:
            break

        yield y.item(), logprobs
        if n % 256 == 0:
            mx.clear_cache()
        y, logprobs = next_y, next_logprobs
        n += 1


def stream_generate(
    model: nn.Module,
    processor: PreTrainedTokenizer,
    prompt: str,
    image: Union[str, List[str]] = None,
    audio: Union[str, List[str]] = None,
    **kwargs,
) -> Union[str, Generator[str, None, None]]:
    """
    A generator producing text based on the given prompt from the model.

    Args:
        model (nn.Module): The model to use for generation.
        processor (PreTrainedTokenizer): The tokenizer/processor.
        prompt (str): The input prompt text.
        image (Union[str, List[str]], optional): Image path(s) or URL(s).
        audio (Union[str, List[str]], optional): Audio file path(s).
        prefill_step_size (int, optional): Number of tokens to process per prefill
          step. When set, enables chunked prefill which processes long prompts in
          smaller chunks to reduce peak memory usage.
        kwargs: Additional options passed to :func:`generate_step`.
          See :func:`generate_step` for more details.

    Yields:
        Generator[GenerationResult]: A generator producing GenerationResult objects
          containing the generated text, tokens, and statistics.
    """
    tokenizer = processor.tokenizer if hasattr(processor, "tokenizer") else processor

    # Skip special tokens
    skip_special_tokens = kwargs.pop("skip_special_tokens", False)
    skip_special_token_ids = (
        set(tokenizer.all_special_ids)
        if skip_special_tokens and hasattr(tokenizer, "all_special_ids")
        else []
    )

    add_special_tokens = (
        not hasattr(processor, "chat_template")
        if model.config.model_type in ["gemma3", "gemma3n"]
        else True
    )

    resize_shape = kwargs.pop("resize_shape", None)
    image_token_index = getattr(model.config, "image_token_index", None)

    if kwargs.get("input_ids", None) is not None:
        input_ids = kwargs.pop("input_ids")
        pixel_values = kwargs.pop("pixel_values", None)
        mask = kwargs.pop("mask", None)
    else:
        inputs = prepare_inputs(
            processor,
            images=image,
            audio=audio,
            prompts=prompt,
            image_token_index=image_token_index,
            resize_shape=resize_shape,
            add_special_tokens=add_special_tokens,
            **kwargs,
        )
        input_ids = inputs.get("input_ids", None)
        pixel_values = inputs.get("pixel_values", None)
        mask = inputs.get("attention_mask", None)
        data_kwargs = {
            k: v
            for k, v in inputs.items()
            if k not in ["input_ids", "pixel_values", "attention_mask"]
        }
        kwargs.update(data_kwargs)

    with wired_limit(model, [generation_stream]):
        detokenizer = processor.detokenizer
        detokenizer.reset()
        tic = time.perf_counter()
        for n, (token, logprobs) in enumerate(
            generate_step(input_ids, model, pixel_values, mask, **kwargs)
        ):
            if n == 0:
                prompt_time = time.perf_counter() - tic
                prompt_tps = input_ids.size / prompt_time
                tic = time.perf_counter()

            # Stop generation if the token is in the eos_token_ids
            if tokenizer.stopping_criteria(token):
                break

            detokenizer.add_token(token, skip_special_token_ids=skip_special_token_ids)

            # Yield the last segment if streaming
            yield GenerationResult(
                text=detokenizer.last_segment,
                token=token,
                logprobs=logprobs,
                prompt_tokens=input_ids.size,
                generation_tokens=n + 1,
                total_tokens=input_ids.size + n + 1,
                prompt_tps=prompt_tps,
                generation_tps=(n + 1) / (time.perf_counter() - tic),
                peak_memory=mx.get_peak_memory() / 1e9,
            )

        detokenizer.finalize()
        yield GenerationResult(
            text=detokenizer.last_segment,
            token=token,
            logprobs=logprobs,
            prompt_tokens=input_ids.size,
            generation_tokens=n + 1,
            total_tokens=input_ids.size + n + 1,
            prompt_tps=prompt_tps,
            generation_tps=(n + 1) / (time.perf_counter() - tic),
            peak_memory=mx.get_peak_memory() / 1e9,
        )

        # Cleanup after generation
        mx.clear_cache()


def generate(
    model: nn.Module,
    processor: PreTrainedTokenizer,
    prompt: str,
    image: Union[str, List[str]] = None,
    audio: Union[str, List[str]] = None,
    verbose: bool = False,
    **kwargs,
) -> GenerationResult:
    """
    Generate text from the model.

    Args:
       model (nn.Module): The language model.
       tokenizer (PreTrainedTokenizer): The tokenizer.
       prompt (str): The string prompt.
       temperature (float): The temperature for sampling (default 0).
       max_tokens (int): The maximum number of tokens (default 100).
       verbose (bool): If ``True``, print tokens and timing information
           (default ``False``).
       formatter (Optional[Callable]): A function which takes a token and a
           probability and displays it.
       repetition_penalty (float, optional): The penalty factor for repeating tokens.
       repetition_context_size (int, optional): The number of tokens to consider for repetition penalty.
    """

    if verbose:
        print("=" * 10)
        files = []
        if image is not None:
            files.extend(image)
        if audio is not None:
            files.extend(audio)
        if kwargs.get("video") is not None:
            files.extend(kwargs.get("video"))

        print(f"Files: {files}", "\n")

        print("Prompt:", prompt)

    text = ""
    last_response = None

    eos_tokens = kwargs.get("eos_tokens", None)
    stopping_criteria = kwargs.get("stopping_criteria", None)

    # Get the tokenizer
    tokenizer = processor.tokenizer if hasattr(processor, "tokenizer") else processor

    # Add custom EOS tokens to the stopping criteria
    if eos_tokens is not None:
        tokenizer.stopping_criteria.add_eos_token_ids(eos_tokens)

    # Use custom stopping criteria
    elif stopping_criteria is not None:
        if isinstance(stopping_criteria, StoppingCriteria) or callable(
            stopping_criteria
        ):
            tokenizer.stopping_criteria = stopping_criteria
        else:
            raise ValueError(
                "stopping_criteria must be an instance of StoppingCriteria or a callable"
            )
    else:
        tokenizer.stopping_criteria.reset(model.config.eos_token_id)

    for response in stream_generate(model, processor, prompt, image, audio, **kwargs):
        if verbose:
            print(response.text, end="", flush=True)
        text += response.text
        last_response = response

    if verbose:
        print("\n" + "=" * 10)
        if len(text) == 0:
            print("No text generated for this prompt")
            return GenerationResult(
                text=text,
                token=None,
                logprobs=None,
                prompt_tokens=0,
                generation_tokens=0,
                total_tokens=0,
                prompt_tps=0.0,
                generation_tps=0.0,
                peak_memory=mx.get_peak_memory() / 1e9,
            )
        print(
            f"Prompt: {last_response.prompt_tokens} tokens, "
            f"{last_response.prompt_tps:.3f} tokens-per-sec"
        )
        print(
            f"Generation: {last_response.generation_tokens} tokens, "
            f"{last_response.generation_tps:.3f} tokens-per-sec"
        )
        print(f"Peak memory: {last_response.peak_memory:.3f} GB")

    return GenerationResult(
        text=text,
        token=last_response.token,
        logprobs=last_response.logprobs,
        prompt_tokens=last_response.prompt_tokens,
        generation_tokens=last_response.generation_tokens,
        total_tokens=last_response.total_tokens,
        prompt_tps=last_response.prompt_tps,
        generation_tps=last_response.generation_tps,
        peak_memory=last_response.peak_memory,
    )


@dataclass
class BatchGenerationResult:
    """
    Result of batch generation with optional image size tracking.

    Attributes:
        texts: Generated text for each sample
        tokens: Last generated token for each sample
        logprobs: Log probabilities for each sample
        prompt_tokens: Number of prompt tokens per sample
        generation_tokens: Number of generated tokens per sample
        total_tokens: Total tokens (prompt + generation) per sample
        prompt_tps: Prompt tokens per second per sample
        generation_tps: Generation tokens per second per sample
        peak_memory: Peak memory usage in GB
        image_sizes: Original (height, width) for each image (for tracking)
    """

    texts: List[str]
    tokens: List[Optional[int]]
    logprobs: List[Optional[List[float]]]
    prompt_tokens: List[int]
    generation_tokens: List[int]
    total_tokens: List[int]
    prompt_tps: List[float]
    generation_tps: List[float]
    peak_memory: float = 0.0
    image_sizes: Optional[List[Tuple[int, int]]] = None


def _left_pad_prompts(prompts, max_length=None):
    if max_length is None:
        max_length = max(len(p) for p in prompts)

    return mx.array([[0] * (max_length - len(p)) + p for p in prompts])


def _make_cache(model, left_padding):
    """
    Convert a list of regular caches into their corresponding
    batch-aware caches.
    """

    def to_batch_cache(c):
        if isinstance(c, cache.KVCache):
            return cache.BatchKVCache(left_padding)
        elif isinstance(c, cache.ArraysCache):
            c.left_padding = mx.array(left_padding)
            return c
        elif isinstance(c, cache.RotatingKVCache):
            if c.keep > 0:
                raise ValueError("RotatingKVCache with keep tokens is not supported.")
            return cache.BatchRotatingKVCache(c.max_size, left_padding)
        elif isinstance(c, cache.CacheList):
            return cache.BatchCacheList(*(to_batch_cache(sub_c) for sub_c in c.caches))
        else:
            raise ValueError(f"{type(c)} does not yet support batching")

    if hasattr(model, "make_cache"):
        model_cache = model.make_cache()
        return [to_batch_cache(c) for c in model_cache]
    else:
        return [cache.BatchKVCache(left_padding) for _ in model.layers]


@dataclass
class BatchStats:
    """
    An data object to hold generation stats.

    Args:
        prompt_tokens (int): The number of prompt tokens processed.
        prompt_tps (float): The prompt processing tokens-per-second.
        prompt_time (float): The time in seconds spent in prompt processing.
        generation_tokens (int): The number of generated tokens.
        generation_tps (float): The tokens-per-second for generation.
        generation_time (float): The time in seconds spent in generation .
        peak_memory (float): The peak memory used so far in GB.
    """

    prompt_tokens: int = 0
    prompt_tps: float = 0
    prompt_time: float = 0
    generation_tokens: int = 0
    generation_tps: float = 0
    generation_time: float = 0
    peak_memory: float = 0


@dataclass
class BatchResponse:
    """
    An data object to hold a batch generation response.

    Args:
        texts: (List[str]): The generated text for each prompt.
        stats (BatchStats): Statistics about the generation.
        image_sizes: (Optional[List[Tuple[int, int]]]): Original (height, width)
            for each image. Useful for tracking which images produced which responses
            and for debugging padding/batching behavior.
    """

    texts: List[str]
    stats: BatchStats
    image_sizes: Optional[List[Tuple[int, int]]] = None


@dataclass
class Batch:
    uids: List[int]
    y: mx.array
    logprobs: mx.array
    max_tokens: List[int]
    num_tokens: List[int]
    cache: List[Any]

    def __len__(self):
        return len(self.uids)

    def filter(self, keep_idx: List[int]):
        self.uids = [self.uids[k] for k in keep_idx]
        self.max_tokens = [self.max_tokens[k] for k in keep_idx]
        self.num_tokens = [self.num_tokens[k] for k in keep_idx]
        keep_idx = mx.array(keep_idx, mx.int32)
        self.y = self.y[keep_idx]
        self.logprobs = self.logprobs[keep_idx]
        for c in self.cache:
            c.filter(keep_idx)

    def extend(self, other):
        self.uids.extend(other.uids)
        self.y = mx.concatenate([self.y, other.y])
        self.logprobs = mx.concatenate([self.logprobs, other.logprobs])
        self.num_tokens.extend(other.num_tokens)
        self.max_tokens.extend(other.max_tokens)
        for c, o in zip(self.cache, other.cache):
            c.extend(o)


class BatchGenerator:

    @dataclass
    class Response:
        uid: int
        token: int
        logprobs: mx.array
        finish_reason: Optional[str]

    def __init__(
        self,
        model,
        processor,
        max_tokens: int = 128,
        stop_tokens: Optional[set] = None,
        sampler: Optional[Callable[[mx.array], mx.array]] = None,
        completion_batch_size: int = 32,
        prefill_batch_size: int = 8,
        prefill_step_size: int = 2048,
        prompt_cache=None,
    ):
        self.model = model
        self.unprocessed_prompts = []
        self.max_tokens = max_tokens
        self.processor = processor
        self.tokenizer = (
            processor.tokenizer if hasattr(processor, "tokenizer") else processor
        )
        self.sampler = sampler or (lambda x: mx.argmax(x, axis=-1))
        self.uid_count = 0
        self.prefill_step_size = prefill_step_size
        self.prefill_batch_size = prefill_batch_size
        self.completion_batch_size = completion_batch_size
        self.prompt_cache = prompt_cache
        self._stats = BatchStats()

        self.tokenizer.stopping_criteria.add_eos_token_ids(stop_tokens)

        self.active_batch = None

    def insert(self, prompts, max_tokens: Union[List[int], int, None] = None):
        uids = []

        if max_tokens is None or isinstance(max_tokens, int):
            max_tokens = [max_tokens or self.max_tokens] * len(prompts)

        for p, m in zip(prompts, max_tokens):
            self.unprocessed_prompts.append((self.uid_count, p, m))
            uids.append(self.uid_count)
            self.uid_count += 1
        # Sort in ascending order of length
        self.unprocessed_prompts = sorted(
            self.unprocessed_prompts, key=lambda x: len(x[1])
        )
        return uids

    def _process_prompts(self, prompts, **kwargs) -> Batch:
        uids, inputs, max_tokens = zip(*prompts)
        lengths = [len(p) for p in inputs]
        max_length = max(lengths)

        self._stats.prompt_tokens += sum(lengths)
        left_padding = [max_length - l for l in lengths]
        inputs = _left_pad_prompts(inputs, max_length=max_length)

        prompt_cache = (
            _make_cache(self.model, left_padding)
            if self.prompt_cache is None
            else self.prompt_cache
        )

        # Slice batch data in kwargs to match current batch size
        batch_size = len(uids)
        for key, value in kwargs.items():
            if isinstance(value, mx.array) and value.ndim > 0:
                kwargs[key] = value[:batch_size]

        inputs_embeds = kwargs.pop("inputs_embeds", None)

        if inputs_embeds is not None:
            # Multimodal prefill
            while inputs_embeds.shape[1] > 1:
                n_to_process = min(self.prefill_step_size, inputs_embeds.shape[1] - 1)
                self.model(
                    inputs[:, :n_to_process],
                    cache=prompt_cache,
                    inputs_embeds=inputs_embeds[:, :n_to_process],
                    n_to_process=n_to_process,
                    **kwargs,
                )
                mx.eval([c.state for c in prompt_cache])
                inputs_embeds = inputs_embeds[:, n_to_process:]
                inputs = inputs[:, n_to_process:]
                mx.clear_cache()

            kwargs = {"inputs_embeds": inputs_embeds}

        else:
            # Text-only prefill
            while inputs.shape[1] > 1 and inputs_embeds is None:
                n_to_process = min(self.prefill_step_size, inputs.shape[1] - 1)
                self.model(inputs[:, :n_to_process], cache=prompt_cache)
                mx.eval([c.state for c in prompt_cache])
                inputs = inputs[:, n_to_process:]
                mx.clear_cache()

        y, logprobs = self._step(inputs, prompt_cache, **kwargs)
        mx.async_eval(y, logprobs)
        mx.clear_cache()
        return Batch(
            list(uids), y, logprobs, list(max_tokens), [0] * len(uids), prompt_cache
        )

    def _step(self, input_tokens: mx.array, prompt_cache: List[Any], **kwargs):
        output = self.model(input_tokens, cache=prompt_cache, **kwargs)
        logits = output.logits[:, -1, :]
        logprobs = logits - mx.logsumexp(logits, axis=-1, keepdims=True)
        sampled = self.sampler(logprobs)

        # TODO: Add KV cache quantization if specified
        return sampled, logprobs

    def stats(self):
        self._stats.prompt_tps = self._stats.prompt_tokens / self._stats.prompt_time
        self._stats.generation_tps = (
            self._stats.generation_tokens / self._stats.generation_time
        )
        self._stats.peak_memory = mx.get_peak_memory() / 1e9
        return self._stats

    def _next(self, **kwargs):
        tic = time.perf_counter()

        prompt_processing = False
        batch = self.active_batch
        num_active = len(batch) if batch else 0
        num_to_add = self.completion_batch_size - num_active
        while num_to_add >= self.prefill_batch_size:
            prompts = self.unprocessed_prompts[: self.prefill_batch_size]
            # Finish processing the last examples of the last batch
            if len(prompts) == 0 and num_active > 0:
                break
            # No more prompts and no more completions, all done
            elif len(prompts) == 0:
                self.active_batch = None
                return []
            # Process prompts
            if batch is not None and not prompt_processing:
                # Finish any active completion tokens
                mx.eval(batch.y, batch.logprobs)
                self._stats.generation_time += time.perf_counter() - tic
                tic = time.perf_counter()

            batch = self._process_prompts(prompts, **kwargs)
            self.unprocessed_prompts = self.unprocessed_prompts[
                self.prefill_batch_size :
            ]
            prompt_processing = True
            # If there was no active batch, set it
            if self.active_batch is None:
                self.active_batch = batch
            else:
                self.active_batch.extend(batch)

            num_active = len(self.active_batch)
            num_to_add -= len(batch)

        batch = self.active_batch
        y, logprobs = batch.y, batch.logprobs
        batch.y, batch.logprobs = self._step(y[:, None], batch.cache)
        mx.async_eval(batch.y, batch.logprobs)

        y = y.tolist()
        toc = time.perf_counter()
        if prompt_processing:
            self._stats.prompt_time += toc - tic
        else:
            self._stats.generation_time += toc - tic
        keep_idx = []
        end_idx = []
        responses = []

        for e, (t, uid, num_tok, max_tok) in enumerate(
            zip(y, batch.uids, batch.num_tokens, batch.max_tokens)
        ):
            num_tok += 1
            batch.num_tokens[e] = num_tok
            if self.tokenizer.stopping_criteria(t):
                finish_reason = "stop"
                end_idx.append(e)
            elif num_tok >= max_tok:
                finish_reason = "length"
                end_idx.append(e)
            else:
                finish_reason = None
                keep_idx.append(e)
            responses.append(self.Response(uid, t, logprobs[e], finish_reason))

        # Remove any finished completions
        if len(end_idx):
            if len(keep_idx) > 0:
                batch.filter(keep_idx)
            else:
                self.active_batch = None

        self._stats.generation_tokens += len(responses)

        if len(responses) > 0 and self._stats.generation_tokens % 100 == 0:
            mx.clear_cache()

        return responses

    def next(self, **kwargs):
        with mx.stream(generation_stream):
            return self._next(**kwargs)


def batch_generate(
    model,
    processor,
    images: Union[str, List[str]] = None,
    audios: Union[str, List[str]] = None,
    prompts: List[str] = None,
    max_tokens: Union[int, List[int]] = 128,
    verbose: bool = False,
    group_by_shape: bool = True,
    track_image_sizes: bool = True,
    **kwargs,
):
    """
    Generate responses for the given batch of prompts with variable-sized images.

    This function implements the transformers-style approach to batching:
    1. Group images with the same shape for efficient batch processing
    2. Process each group as a batch (no padding waste within groups)
    3. Track original image sizes for proper attention masking
    4. Restore results to original batch order

    Key insight: Instead of padding all images to the same spatial dimensions
    (which wastes computation and may hurt accuracy), we group same-sized
    images together so there's zero padding within each group.

    Args:
       model (nn.Module): The language model.
       processor (PreTrainedTokenizer): The tokenizer/processor.
       images (Union[str, List[str]]): Images (paths, URLs, or PIL images).
       audios (Union[str, List[str]]): Audio files (not yet supported for batching).
       prompts (List[str]): The input prompts.
       max_tokens (Union[int, List[int]]): Maximum number of output tokens. This
          can be per prompt if a list is provided.
       verbose (bool): If ``True``, print tokens and timing information.
          Default: ``False``.
       group_by_shape (bool): If ``True``, group same-shaped images for efficient
          batch processing. Default: ``True``.
       track_image_sizes (bool): If ``True``, track and return original image sizes.
          Default: ``True``.
       kwargs: The remaining options get passed to :obj:`BatchGenerator`.
          See :obj:`BatchGenerator` for more details.

    Returns:
        BatchResponse with generated texts, statistics, and optionally image_sizes.
    """
    from PIL import Image

    from .utils import process_image

    processor.detokenizer.reset()
    tokenizer = processor.tokenizer if hasattr(processor, "tokenizer") else processor

    # Handle single image case
    if isinstance(images, str):
        images = [images]

    # Handle no images case
    if images is None:
        texts, stats = _generate_batch(
            model, processor, prompts, None, max_tokens, verbose, **kwargs
        )
        return BatchResponse(texts, stats)

    # Load and preprocess images
    image_processor = (
        processor.image_processor if hasattr(processor, "image_processor") else None
    )

    processed_images = []
    image_sizes_original = []
    for img in images:
        if isinstance(img, str):
            pil_img = process_image(img, None, image_processor)
        elif isinstance(img, Image.Image):
            pil_img = img
        else:
            pil_img = img
        processed_images.append(pil_img)
        # Track original size
        if hasattr(pil_img, "height"):
            image_sizes_original.append((pil_img.height, pil_img.width))
        else:
            image_sizes_original.append((0, 0))

    # Group images by shape for efficient processing (no padding within groups)
    if group_by_shape and len(processed_images) > 1:
        grouped_images, grouped_indices = group_images_by_shape(processed_images)

        if verbose:
            print(f"[batch_generate] Found {len(grouped_images)} unique image shapes")
    else:
        # Single image or grouping disabled - treat as one group
        shape = (
            (processed_images[0].height, processed_images[0].width)
            if processed_images
            else (0, 0)
        )
        grouped_images = {shape: processed_images}
        grouped_indices = {shape: list(range(len(processed_images)))}

    # Process each shape group
    all_texts = [None] * len(prompts)
    all_image_sizes = [None] * len(prompts)
    total_stats = BatchStats()

    for shape, indices in grouped_indices.items():
        # Get images and prompts for this shape group
        group_images = [processed_images[i] for i in indices]
        group_prompts = [prompts[i] for i in indices]
        group_sizes = [image_sizes_original[i] for i in indices]

        # Handle per-sample max_tokens
        if isinstance(max_tokens, list):
            group_max_tokens = [max_tokens[i] for i in indices]
        else:
            group_max_tokens = max_tokens

        # Process the entire group at once (same shape = no padding needed)
        chunk_texts, chunk_stats = _generate_batch(
            model,
            processor,
            group_prompts,
            group_images,
            group_max_tokens,
            **kwargs,
        )

        # Store results in original order
        for j, orig_idx in enumerate(indices):
            all_texts[orig_idx] = chunk_texts[j]
            all_image_sizes[orig_idx] = group_sizes[j]

        # Accumulate stats
        total_stats.prompt_tokens += chunk_stats.prompt_tokens
        total_stats.prompt_time += chunk_stats.prompt_time
        total_stats.generation_tokens += chunk_stats.generation_tokens
        total_stats.generation_time += chunk_stats.generation_time

    mx.clear_cache()

    # Compute final stats
    if total_stats.prompt_time > 0:
        total_stats.prompt_tps = total_stats.prompt_tokens / total_stats.prompt_time
    if total_stats.generation_time > 0:
        total_stats.generation_tps = (
            total_stats.generation_tokens / total_stats.generation_time
        )
    total_stats.peak_memory = mx.get_peak_memory() / 1e9

    if verbose:
        print(f"[batch_generate] Finished processing {len(prompts)} samples")
        print(
            f"[batch_generate] Prompt: {total_stats.prompt_tokens} tokens, {total_stats.prompt_tps:.3f} tokens-per-sec"
        )
        print(
            f"[batch_generate] Generation: {total_stats.generation_tokens} tokens, "
            f"{total_stats.generation_tps:.3f} tokens-per-sec"
        )
        print(f"[batch_generate] Peak memory: {total_stats.peak_memory:.3f} GB")

    response = BatchResponse(all_texts, total_stats)
    if track_image_sizes:
        response.image_sizes = all_image_sizes
    return response


def _generate_batch(
    model,
    processor,
    prompts: List[str],
    images: List = None,
    max_tokens: Union[int, List[int]] = 100,
    verbose: bool = False,
    **kwargs,
) -> Tuple[List[str], BatchStats]:

    tokenizer = processor.tokenizer if hasattr(processor, "tokenizer") else processor
    batch_size = len(prompts)

    num_images_list = [
        1 if i < (len(images) if images is not None else 0) else 0
        for i in range(len(prompts))
    ]
    formatted_prompts = [
        apply_chat_template(
            processor,
            model.config,
            p,
            num_images=num_images_list[i],
        )
        for i, p in enumerate(prompts)
    ]

    add_special_tokens = (
        not hasattr(processor, "chat_template")
        if model.config.model_type in ["gemma3", "gemma3n"]
        else True
    )

    resize_shape = kwargs.pop("resize_shape", None)
    image_token_index = getattr(model.config, "image_token_index", None)

    inputs = prepare_inputs(
        processor,
        images=images,
        audio=None,
        prompts=formatted_prompts,
        image_token_index=image_token_index,
        resize_shape=resize_shape,
        add_special_tokens=add_special_tokens,
        pad_to_uniform_size=False,  # Since images are pre-grouped by shape, they're already uniform size
    )
    input_ids = inputs.get("input_ids", None)
    pixel_values = inputs.get("pixel_values", None)

    data_kwargs = {
        k: v
        for k, v in inputs.items()
        if k not in ["input_ids", "pixel_values", "attention_mask"]
    }

    # Use batch_size for prefill and completion to ensure consistent processing
    gen = BatchGenerator(
        model.language_model,
        processor,
        prefill_batch_size=batch_size,
        completion_batch_size=batch_size,
        **kwargs,
    )

    with wired_limit(model, [generation_stream]):
        if pixel_values is not None:
            embedding_output = model.get_input_embeddings(
                input_ids, pixel_values, **data_kwargs
            )

            # Normalize embedding output to a kwargs dict expected by BatchGenerator
            if isinstance(embedding_output, dict):
                embed_kwargs = embedding_output
            elif hasattr(embedding_output, "to_dict"):
                # Convert to dict and keep non-None fields
                embed_kwargs = {
                    k: v for k, v in embedding_output.to_dict().items() if v is not None
                }
            else:
                # Assume it's directly an inputs_embeds array
                embed_kwargs = {"inputs_embeds": embedding_output}

            gen_kwargs = {
                "pixel_values": pixel_values,
                **data_kwargs,
                **embed_kwargs,
            }
        else:
            input_ids = mx.squeeze(input_ids, axis=0)
            gen_kwargs = {}

        uids = gen.insert(input_ids.tolist(), max_tokens)
        results = {uid: [] for uid in uids}
        while responses := gen.next(**gen_kwargs):
            for r in responses:
                if r.finish_reason != "stop":
                    results[r.uid].append(r.token)

    texts = [tokenizer.decode(results[uid]) for uid in uids]
    return texts, gen.stats()


def main():
    args = parse_arguments()
    if isinstance(args.image, str):
        args.image = [args.image]

    model, processor = load(
        args.model,
        args.adapter_path,
        revision=args.revision,
        trust_remote_code=args.trust_remote_code,
        quantize_activations=args.quantize_activations,
    )
    config = model.config

    prompt = args.prompt

    num_images = len(args.image) if args.image is not None else 0
    num_audios = (
        1 if args.audio is not None else 0
    )  # TODO: Support multiple audio files
    prompt = apply_chat_template(
        processor, config, prompt, num_images=num_images, num_audios=num_audios
    )

    kwargs = {}

    if args.resize_shape is not None:
        if len(args.resize_shape) not in [1, 2]:
            raise ValueError("Resize shape must be 1 or 2 integers")
        kwargs["resize_shape"] = (
            (args.resize_shape[0],) * 2
            if len(args.resize_shape) == 1
            else tuple(args.resize_shape)
        )

    if args.eos_tokens is not None:
        eos_tokens = []
        for token in args.eos_tokens:
            try:
                decoded_token = codecs.decode(token, "unicode_escape")
                eos_tokens.append(decoded_token)
            except (UnicodeDecodeError, UnicodeError):
                eos_tokens.append(token)
        kwargs["eos_tokens"] = eos_tokens

    if args.skip_special_tokens:
        kwargs["skip_special_tokens"] = args.skip_special_tokens

    # Add processor kwargs from JSON
    if args.processor_kwargs:
        kwargs.update(args.processor_kwargs)

    if args.chat:
        chat = []
        if args.system:
            chat.append({"role": "system", "content": args.system})
        while user := input("User:"):
            chat.append({"role": "user", "content": user})
            prompt = apply_chat_template(processor, config, chat, num_images=num_images)
            response = ""
            print("Assistant:", end="")
            stream_kwargs = {
                "max_tokens": args.max_tokens,
                "temperature": args.temperature,
                **kwargs,
            }
            if args.prefill_step_size is not None:
                stream_kwargs["prefill_step_size"] = args.prefill_step_size

            for chunk in stream_generate(
                model,
                processor,
                prompt,
                args.image,
                args.audio,
                **stream_kwargs,
            ):
                response += chunk.text
                print(chunk.text, end="")

            chat.append({"role": "assistant", "content": response})
            print()

    else:
        gen_kwargs = {
            "image": args.image,
            "audio": args.audio,
            "temperature": args.temperature,
            "max_tokens": args.max_tokens,
            "verbose": args.verbose,
            "max_kv_size": args.max_kv_size,
            "kv_bits": args.kv_bits,
            "kv_group_size": args.kv_group_size,
            "quantized_kv_start": args.quantized_kv_start,
            **kwargs,
        }
        if args.prefill_step_size is not None:
            gen_kwargs["prefill_step_size"] = args.prefill_step_size

        result = generate(
            model,
            processor,
            prompt,
            **gen_kwargs,
        )
        if not args.verbose:
            print(result.text)


if __name__ == "__main__":
    print(
        "Calling `python -m mlx_vlm.generate ...` directly is deprecated."
        " Use `mlx_vlm generate` or `python -m mlx_vlm generate` instead."
    )
    main()



import mlx.core as mx
import mlx.nn as nn

from ..qwen3_5 import Model as Qwen3_5Model
from .config import ModelConfig
from .language import LanguageModel
from .vision import VisionModel


class Model(Qwen3_5Model):

    def __init__(self, config: ModelConfig):
        # only initialize nn.Module, skip the initialization of vision_tower and language_model in the parent class
        nn.Module.__init__(self)
        self.config = config
        self.vision_tower = VisionModel(config.vision_config)
        self.language_model = LanguageModel(config.text_config, config)

    def sanitize(self, weights):
        # ignore mtp weights
        weights = {key: value for key, value in weights.items() if "mtp." not in key}

        if self.config.text_config.tie_word_embeddings:
            weights.pop("lm_head.weight", None)

        for l in range(self.config.text_config.num_hidden_layers):
            prefix = f"model.language_model.layers.{l}.mlp"
            # process gate_up_proj [num_experts, 2 * intermediate_size, hidden_size]
            gate_up_weight = weights.pop(f"{prefix}.experts.gate_up_proj")
            gate_weight, up_weights = mx.split(gate_up_weight, 2, axis=-2)
            weights[f"{prefix}.switch_mlp.gate_proj.weight"] = gate_weight
            weights[f"{prefix}.switch_mlp.up_proj.weight"] = up_weights
            # down_proj
            weights[f"{prefix}.switch_mlp.down_proj.weight"] = weights.pop(
                f"{prefix}.experts.down_proj"
            )

        norm_keys = (
            ".input_layernorm.weight",
            ".post_attention_layernorm.weight",
            "model.norm.weight",
            ".q_norm.weight",
            ".k_norm.weight",
        )

        sanitized_weights = {}
        for key, value in weights.items():
            if "model" in key:
                if "model.language_model" in key:
                    key = key.replace("model.language_model", "language_model.model")
                elif "model.visual" in key:
                    key = key.replace("model.visual", "vision_tower")
            elif "lm_head" in key:
                key = key.replace("lm_head", "language_model.lm_head")

            if "conv1d.weight" in key and value.shape[-1] != 1:
                value = value.moveaxis(2, 1)
            if any(key.endswith(sfx) for sfx in norm_keys):
                if value.ndim == 1:
                    value += 1.0

            sanitized_weights[key] = value

        return sanitized_weights



from typing import Optional

import mlx.core as mx
import mlx.nn as nn
import numpy as np
from transformers import AutoProcessor

from ..base import InputEmbeddingsFeatures
from ..deepseekocr.language import LanguageModel
from ..deepseekocr.sam import SAMEncoder
from .config import ModelConfig, SAMViTConfig
from .processing_deepseekocr import DeepseekOCR2Processor
from .vision import VisionModel

AutoProcessor.register("deepseekocr_2", DeepseekOCR2Processor)


class MlpProjector(nn.Module):
    def __init__(self, config: ModelConfig):
        super().__init__()
        self.config = config

        if config.projector_config.projector_type == "linear":
            self.layers = nn.Linear(
                config.projector_config.input_dim, config.projector_config.n_embed
            )
        else:
            raise ValueError(
                f"Unknown projector type: {config.projector_config.projector_type}"
            )

    def __call__(self, x):
        return self.layers(x)


class Model(nn.Module):
    def __init__(self, config: ModelConfig):
        super().__init__()
        self.config = config
        self.vision_model = VisionModel(config.vision_config)
        sam_config = SAMViTConfig()
        self.sam_model = SAMEncoder(
            img_size=sam_config.image_size,
            patch_size=sam_config.patch_size,
            embed_dim=sam_config.width,
            depth=sam_config.layers,
            num_heads=sam_config.heads,
            window_size=sam_config.window_size,
            global_attn_indexes=sam_config.global_attn_indexes,
            final_out_chans=896,  # OCR-2 uses 896 output channels (vs 1024 in OCR)
        )
        self.language_model = LanguageModel(config.text_config)
        self.projector = MlpProjector(config)

        self.tile_tag = config.tile_tag
        self.global_view_pos = config.global_view_pos

        # view_separator is loaded from model weights (mapped from view_seperator)
        # Initialize with zeros - will be overwritten when weights are loaded
        if self.tile_tag == "2D":
            # <|view_separator|> - marks end of image features
            # Note: This must be defined as an mx.array for weight loading to work
            self.view_separator = mx.zeros((config.projector_config.n_embed,))
        else:
            raise ValueError(
                f"Only 2D tile_tag is supported currently, got: {self.tile_tag}"
            )

    def get_input_embeddings(
        self,
        input_ids: Optional[mx.array] = None,
        pixel_values: Optional[mx.array] = None,
        images_spatial_crop: Optional[mx.array] = None,
        images_seq_mask: Optional[mx.array] = None,
        **kwargs,
    ):
        input_embeds = self.language_model.model.embed_tokens(input_ids)

        if pixel_values is None:
            return InputEmbeddingsFeatures(inputs_embeds=input_embeds)

        # pixel_values is a list: [patches, global_images]
        if isinstance(pixel_values, list):
            patches, global_images = pixel_values
        else:
            patches = None
            global_images = pixel_values

        # Check if we have valid pixel values
        if mx.sum(global_images).item() == 0:
            return InputEmbeddingsFeatures(inputs_embeds=input_embeds)

        # Process images through SAM -> Qwen2 -> Projector pipeline
        batch_size = input_ids.shape[0]

        for idx in range(batch_size):
            all_features = []

            # Check if we have valid patches (non-zero)
            has_patches = patches is not None and mx.sum(patches).item() != 0

            if has_patches:
                # Get spatial crop info for this batch item
                if (
                    images_spatial_crop is not None
                    and idx < images_spatial_crop.shape[0]
                ):
                    rows, cols = int(images_spatial_crop[idx, 0].item()), int(
                        images_spatial_crop[idx, 1].item()
                    )
                    num_patches = rows * cols
                else:
                    num_patches = patches.shape[0]

                # Process each patch through SAM -> Qwen2 -> Projector
                # patches shape: (num_patches, C, H, W) where H=W=768
                for patch_idx in range(num_patches):
                    if patch_idx >= patches.shape[0]:
                        break

                    patch = patches[patch_idx : patch_idx + 1]  # (1, C, H, W)

                    # Transpose to (B, H, W, C) for MLX conv2d
                    patch_hwc = patch.transpose(0, 2, 3, 1)

                    # SAM encoder: (1, 768, 768, 3) -> (1, 12, 12, 896)
                    sam_features = self.sam_model(patch_hwc)

                    # Qwen2 encoder: (1, 12, 12, 896) -> (1, 144, 896)
                    # Uses query_768 automatically based on 144 input tokens
                    vision_features = self.vision_model(patch_hwc, sam_features)

                    # Linear projector: (1, 144, 896) -> (1, 144, 1280)
                    vision_features = self.projector(vision_features)

                    # Remove batch dimension: (144, 1280)
                    all_features.append(vision_features[0])

            # Process global view through SAM -> Qwen2 -> Projector
            # global_images is (N, C, H, W) where H=W=1024
            global_image = global_images[idx : idx + 1]  # (1, C, H, W)

            # Transpose to (B, H, W, C) for MLX conv2d
            global_hwc = global_image.transpose(0, 2, 3, 1)

            # SAM encoder: (1, 1024, 1024, 3) -> (1, 16, 16, 896)
            sam_features = self.sam_model(global_hwc)

            # Qwen2 encoder: (1, 16, 16, 896) -> (1, 256, 896)
            # Uses query_1024 automatically based on 256 input tokens
            global_features = self.vision_model(global_hwc, sam_features)

            # Linear projector: (1, 256, 896) -> (1, 256, 1280)
            global_features = self.projector(global_features)

            # Remove batch dimension: (256, 1280)
            all_features.append(global_features[0])

            # Add view_separator
            all_features.append(self.view_separator[None, :])

            # Concatenate all features: [local_patches..., global, view_sep]
            # Shape: (num_patches * 144 + 256 + 1, 1280)
            vision_features = mx.concatenate(all_features, axis=0)

            # Find positions where images should be placed
            if images_seq_mask is not None:
                image_indices = np.where(images_seq_mask[idx])[0].tolist()
                # Assign image features to those positions
                if len(image_indices) > 0:
                    num_positions = len(image_indices)
                    if vision_features.shape[0] >= num_positions:
                        input_embeds[idx, image_indices] = vision_features[
                            :num_positions
                        ]
                    else:
                        # If we have fewer features than expected, pad with the last features
                        input_embeds[idx, image_indices[: vision_features.shape[0]]] = (
                            vision_features
                        )

        return InputEmbeddingsFeatures(inputs_embeds=input_embeds)

    @property
    def layers(self):
        return self.language_model.model.layers

    def __call__(
        self,
        input_ids: mx.array,
        pixel_values: Optional[mx.array] = None,
        mask: Optional[mx.array] = None,
        cache=None,
        **kwargs,
    ):
        images_spatial_crop = kwargs.get("images_spatial_crop", None)
        images_seq_mask = kwargs.get("images_seq_mask", None)

        input_embeddings_features = self.get_input_embeddings(
            input_ids, pixel_values, images_spatial_crop, images_seq_mask
        )

        logits = self.language_model(
            input_ids,
            cache=cache,
            inputs_embeds=input_embeddings_features.inputs_embeds,
        )
        return logits

    @staticmethod
    def sanitize(weights):
        def transform_key(key):
            # Handle Qwen2 encoder weights from HuggingFace format
            # HuggingFace: model.qwen2_model.model.model.layers.X...
            # MLX: vision_model.qwen2_encoder.layers.X...
            if "qwen2_model.model.model.layers" in key:
                key = key.replace(
                    "model.qwen2_model.model.model.layers",
                    "vision_model.qwen2_encoder.layers",
                )

            # Handle Qwen2 encoder norm
            if "qwen2_model.model.model.norm" in key:
                key = key.replace(
                    "model.qwen2_model.model.model.norm",
                    "vision_model.qwen2_encoder.norm",
                )

            # Handle query weights (learnable queries for Qwen2 encoder)
            # For 1024x1024 images, SAM outputs 16x16=256 features, so use query_1024
            # query_1024: (256, 896) - used for 1024x1024 images
            # query_768: (144, 896) - used for 768x768 images
            if "model.qwen2_model.query_1024.weight" in key:
                key = key.replace(
                    "model.qwen2_model.query_1024.weight",
                    "vision_model.qwen2_encoder.query_1024",
                )
            elif "model.qwen2_model.query_1024" in key:
                key = key.replace(
                    "model.qwen2_model.query_1024",
                    "vision_model.qwen2_encoder.query_1024",
                )
            # Also handle query_768 for smaller images (not currently used but keep for future)
            if "model.qwen2_model.query_768.weight" in key:
                key = key.replace(
                    "model.qwen2_model.query_768.weight",
                    "vision_model.qwen2_encoder.query_768",
                )
            elif "model.qwen2_model.query_768" in key:
                key = key.replace(
                    "model.qwen2_model.query_768",
                    "vision_model.qwen2_encoder.query_768",
                )

            # Language model layers
            if (
                "model.layers" in key
                and "language_model" not in key
                and "qwen2" not in key
            ):
                key = key.replace("model.layers", "language_model.model.layers")

            if (
                "model.embed_tokens" in key
                and "language_model" not in key
                and "qwen2" not in key
            ):
                key = key.replace(
                    "model.embed_tokens", "language_model.model.embed_tokens"
                )

            if (
                "model.norm" in key
                and "language_model" not in key
                and "qwen2" not in key
            ):
                key = key.replace("model.norm", "language_model.model.norm")

            if "model.vision_model" in key:
                key = key.replace("model.vision_model", "vision_model")

            if "model.sam_model" in key:
                key = key.replace("model.sam_model", "sam_model")

            if "model.projector" in key:
                key = key.replace("model.projector", "projector")

            # Note: HuggingFace has typo "view_seperator" (e instead of a)
            if "model.view_seperator" in key:
                key = key.replace("model.view_seperator", "view_separator")

            if "lm_head.weight" in key and "language_model" not in key:
                key = key.replace("lm_head.weight", "language_model.lm_head.weight")

            return key

        return {transform_key(k): v for k, v in weights.items()}
